# ***************************************************************************
# *   Copyright (c) 2024 Peter McB                                          *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD result mechanical task panel"
__author__ = "PMcB"
__url__ = "https://www.freecad.org"

## @package view_result_mechanical
#  \ingroup FEM
#  \brief Create and Place Labels on the screen with an optional image.

import FreeCADGui
from pivy import coin

"""
Place Label on the screen with an optional image.
positions on screen:
         (-.9, .9, 0)    top left
         (.9, .9, 0)     upper right
         (-.9, -.9, 0)   lower left
         (.9, -.9, 0)    lower right
"""


class createLabel:
    def __init__(self, trans, text, image="") -> None:
        self.textSep = coin.SoSeparator()
        self.cam = coin.SoOrthographicCamera()
        self.cam.aspectRatio = 1
        self.cam.viewportMapping = coin.SoCamera.LEAVE_ALONE

        self.trans = coin.SoTranslation()
        self.trans.translation = trans
        self.translation = trans
        self.color = coin.SoBaseColor()
        self.color.rgb = (0, 0, 0)

        self.myFont = coin.SoFont()
        self.myFont.name = "Arial,FreeSans,sans"
        #       self.myFont.name = "FreeMono,FreeSans,sans"
        self.size = 20  # 24
        self.myFont.size.setValue(self.size)

        self.SoText2 = coin.SoText2()
        self.SoText2.string = text

        self.textSep.addChild(self.cam)
        self.textSep.addChild(self.trans)
        self.textSep.addChild(self.color)
        self.textSep.addChild(self.myFont)
        self.textSep.addChild(self.SoText2)
        self.activeDoc = FreeCADGui.ActiveDocument
        self.view = self.activeDoc.ActiveView
        self.viewer = self.view.getViewer()
        self.render = self.viewer.getSoRenderManager()
        self.isimage = False
        if image != "":
            self.add_image(image)
            self.isimage = True
        self.sup = self.render.addSuperimposition(self.textSep)
        self.displayed = True

    # get position of label
    def get_position(self):
        return self.translation

    # set position of label
    def set_position(self, trans):
        self.trans.translation = trans
        self.translation = trans

    # set font
    def set_font(self, font):
        self.myFont.name = font

    # set font size
    def set_font_size(self, font_size):
        self.myFont.size.setValue(font_size)

    # set text of label
    def set_text(self, text):
        self.SoText2.string = text

    # set colour of label
    def set_colour(self, colour):
        self.color.rgb = colour
        self.textSep.addChild(self.color)

    # add or change the image with the label
    def add_image(self, image):
        self.remove_image()
        self.isimage = True
        self.myImage = coin.SoImage()
        self.myImage.filename.setValue(image)
        self.textSep.addChild(self.myImage)

    # remove the image
    def remove_image(self):
        if self.isimage:
            # self.myImage = coin.SoImage()
            self.textSep.removeChild(self.myImage)
            self.isimage = False

    # display the label
    def display(self):
        if not self.displayed:
            # display the Superimposition layer with :
            self.sup = self.render.addSuperimposition(self.textSep)
            self.displayed = True

    # hide the label
    def hide(self):
        # hide the Superimposition layer with :
        if self.displayed:
            self.render.removeSuperimposition(self.sup)
            self.displayed = False

    # help
    def help(self):
        print(
            """
Place Label on the screen with an optional image.
Create an instance:
import CreateLabels as CL
label =  CL.creatLabel(<position of label>, <title of label>, <optional image>)
e.g.:
label1 = CL.createLabel( (-0.7, -0.90, 0), "this is text")
label2 = CL.createLabel( (+0.5, -0.90, 0), "this is a image", image = "mesh.jpg")
Positions on screen:
        (-.9, .9, 0)    top left
        (.9, .9, 0)     upper right
        (-.9, -.9, 0)   lower left
        (.9, -.9, 0)    lower right

The following function are available:
        label1.set_position((0.7, 0.50, 0))
        pos= label1.get_position()
        label1.set_font("FreeMono,FreeSans,sans")
        label1.set_font_size(20)
        label1.set_text("hello world")
		label1.set_colour((0,1,0)) # colour of text
        label1.add_image("opera.jpg")
        label1.remove_image()
        label1.hide()  - hide the label
        label1.display()  - display the label, after hide
"""
        )


if __name__ == "__main__":
    label1 = createLabel((-0.98, 0.90, 0), "this a new text")
"""
Examples:
import CreateLabels as CL
label1 = CL.createLabel( (+0.5, -0.90, 0), "this is a image", image = "mesh.jpg")
label2 = CL.createLabel( (-0.7, -0.90, 0), "this is text")

label1.set_colour((0,1,0))
label1.remove_image()
label1.add_image("opera.jpg")
label1.set_text("hello world")
label1.set_position((+0.7, 0.50, 0))
label1.set_font("FreeMono,FreeSans,sans")
label1.hide()
"""



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM DrawingView command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class BIM_DrawingView:
    """The command definition for the Drawing View command"""

    def GetResources(self):

        return {
            "Pixmap": "BIM_ArchView",
            "MenuText": QT_TRANSLATE_NOOP("BIM_DrawingView", "2D Drawing"),
            "Accel": "V, D",
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_DrawingView", "Creates a drawing container to contain elements of a 2D view"
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        import Draft

        FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create 2D View"))
        FreeCADGui.addModule("Arch")
        FreeCADGui.addModule("Draft")
        FreeCADGui.addModule("WorkingPlane")
        FreeCADGui.doCommand("obj = Arch.make2DDrawing()")
        FreeCADGui.doCommand("Draft.autogroup(obj)")
        s = FreeCADGui.Selection.getSelection()
        if len(s) == 1:
            s = s[0]
            if Draft.getType(s) == "SectionPlane":
                FreeCADGui.doCommand(
                    "vobj = Draft.make_shape2dview(FreeCAD.ActiveDocument." + s.Name + ")"
                )
                FreeCADGui.doCommand('vobj.Label = "' + translate("BIM", "Viewed lines") + '"')
                FreeCADGui.doCommand("vobj.InPlace = False")
                FreeCADGui.doCommand("obj.addObject(vobj)")
                bb = FreeCAD.BoundBox()
                for so in s.Objects:
                    if hasattr(so, "Shape"):
                        bb.add(so.Shape.BoundBox)
                if bb.isInside(s.Shape.CenterOfMass):
                    FreeCADGui.doCommand(
                        "cobj = Draft.make_shape2dview(FreeCAD.ActiveDocument." + s.Name + ")"
                    )
                    FreeCADGui.doCommand('cobj.Label = "' + translate("BIM", "Cut lines") + '"')
                    FreeCADGui.doCommand("cobj.InPlace = False")
                    FreeCADGui.doCommand('cobj.ProjectionMode = "Cutfaces"')
                    FreeCADGui.doCommand("obj.addObject(cobj)")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


FreeCADGui.addCommand("BIM_DrawingView", BIM_DrawingView())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2022 - 2025 Larry Woestman <LarryWoestman2@gmail.com>   *
# *   Copyright (c) 2024 Ondsel <development@ondsel.com>                    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import argparse

from typing import Any, Dict

from Path.Post.Processor import PostProcessor

import Path
import FreeCAD

translate = FreeCAD.Qt.translate

DEBUG = False
if DEBUG:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

#
# Define some types that are used throughout this file.
#
Defaults = Dict[str, bool]
Values = Dict[str, Any]
Visible = Dict[str, bool]


class Refactored_Grbl(PostProcessor):
    """The Refactored Grbl post processor class."""

    def __init__(
        self,
        job,
        tooltip=translate("CAM", "Refactored Grbl post processor"),
        tooltipargs=[""],
        units="Metric",
    ) -> None:
        super().__init__(
            job=job,
            tooltip=tooltip,
            tooltipargs=tooltipargs,
            units=units,
        )
        Path.Log.debug("Refactored Grbl post processor initialized.")

    def init_values(self, values: Values) -> None:
        """Initialize values that are used throughout the postprocessor."""
        #
        super().init_values(values)
        #
        # Set any values here that need to override the default values set
        # in the parent routine.
        #
        values["ENABLE_COOLANT"] = True
        #
        # If this is set to True, then commands that are placed in
        # comments that look like (MC_RUN_COMMAND: blah) will be output.
        #
        values["ENABLE_MACHINE_SPECIFIC_COMMANDS"] = True
        #
        # Used in the argparser code as the "name" of the postprocessor program.
        # This would normally show up in the usage message in the TOOLTIP_ARGS.
        #
        values["MACHINE_NAME"] = "Grbl"
        #
        # Default to outputting Path labels at the beginning of each Path.
        #
        values["OUTPUT_PATH_LABELS"] = True
        #
        # Default to not outputting M6 tool changes (comment it) as grbl
        # currently does not handle it.
        #
        values["OUTPUT_TOOL_CHANGE"] = False
        #
        # The order of the parameters.
        # Arcs may only work on the XY plane (this needs to be verified).
        #
        values["PARAMETER_ORDER"] = [
            "X",
            "Y",
            "Z",
            "A",
            "B",
            "C",
            "U",
            "V",
            "W",
            "I",
            "J",
            "K",
            "F",
            "S",
            "T",
            "Q",
            "R",
            "L",
            "P",
        ]
        #
        # Any commands in this value will be output as the last commands in the G-code file.
        #
        values[
            "POSTAMBLE"
        ] = """M5
G17 G90
M2"""
        values["POSTPROCESSOR_FILE_NAME"] = __name__
        #
        # Any commands in this value will be output after the header and
        # safety block at the beginning of the G-code file.
        #
        values["PREAMBLE"] = """G17 G90"""
        #
        # Do not show the current machine units just before the PRE_OPERATION.
        #
        values["SHOW_MACHINE_UNITS"] = False
        #
        # Default to not outputting a G43 following tool changes
        #
        values["USE_TLO"] = False

    def init_argument_defaults(self, argument_defaults: Defaults) -> None:
        """Initialize which arguments (in a pair) are shown as the default argument."""
        super().init_argument_defaults(argument_defaults)
        #
        # Modify which argument to show as the default in flag-type arguments here.
        # If the value is True, the first argument will be shown as the default.
        # If the value is False, the second argument will be shown as the default.
        #
        # For example, if you want to show Metric mode as the default, use:
        #   argument_defaults["metric_inch"] = True
        #
        # If you want to show that "Don't pop up editor for writing output" is
        # the default, use:
        #   argument_defaults["show-editor"] = False.
        #
        # Note:  You also need to modify the corresponding entries in the "values" hash
        #        to actually make the default value(s) change to match.
        #
        argument_defaults["tlo"] = False
        argument_defaults["tool_change"] = False

    def init_arguments_visible(self, arguments_visible: Visible) -> None:
        """Initialize which argument pairs are visible in TOOLTIP_ARGS."""
        super().init_arguments_visible(arguments_visible)
        #
        # Modify the visibility of any arguments from the defaults here.
        #
        arguments_visible["bcnc"] = True
        arguments_visible["axis-modal"] = False
        arguments_visible["return-to"] = True
        arguments_visible["tlo"] = False
        arguments_visible["tool_change"] = True
        arguments_visible["translate_drill"] = True
        arguments_visible["wait-for-spindle"] = True

    @property
    def tooltip(self):
        tooltip: str = """
        This is a postprocessor file for the CAM workbench.
        It is used to take a pseudo-gcode fragment from a CAM object
        and output 'real' GCode suitable for a Grbl 3 axis mill.
        """
        return tooltip



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2023 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This module contains geometry editing and geometry properties-related tools"""

import ifcopenshell
import ifcopenshell.util.unit

import FreeCAD

from . import ifc_tools


def add_geom_properties(obj):
    """Adds geometry properties to a FreeCAD object"""

    element = ifc_tools.get_ifc_element(obj)
    if not ifc_tools.has_representation(element):
        return
    ifcfile = ifc_tools.get_ifcfile(obj)
    scaling = ifcopenshell.util.unit.calculate_unit_scale(ifcfile)
    scaling = scaling * 1000  # given scale is for m, we work in mm
    for rep in element.Representation.Representations:
        if rep.RepresentationIdentifier == "Body":
            if len(rep.Items) == 1:
                # Extrusions
                if rep.Items[0].is_a("IfcExtrudedAreaSolid"):
                    ext = rep.Items[0]
                    if "ExtrusionDepth" not in obj.PropertiesList:
                        obj.addProperty(
                            "App::PropertyLength", "ExtrusionDepth", "Geometry", locked=True
                        )
                    obj.ExtrusionDepth = ext.Depth * scaling
                    if "ExtrusionDirection" not in obj.PropertiesList:
                        obj.addProperty(
                            "App::PropertyVector", "ExtrusionDirection", "Geometry", locked=True
                        )
                    obj.ExtrusionDirection = FreeCAD.Vector(ext.ExtrudedDirection.DirectionRatios)

                    # Extrusion of a rectangle
                    if ext.SweptArea.is_a("IfcRectangleProfileDef"):
                        if "RectangleLength" not in obj.PropertiesList:
                            obj.addProperty(
                                "App::PropertyLength", "RectangleLength", "Geometry", locked=True
                            )
                        obj.RectangleLength = ext.SweptArea.XDim * scaling
                        if "RectangleWidth" not in obj.PropertiesList:
                            obj.addProperty(
                                "App::PropertyLength", "RectangleWidth", "Geometry", locked=True
                            )
                        obj.RectangleWidth = ext.SweptArea.YDim * scaling

                    # Extrusion of a polyline
                    elif ext.SweptArea.is_a("IfcArbitraryClosedProfileDef"):
                        if ext.SweptArea.OuterCurve.is_a("IfcPolyline"):
                            if "PolylinePoints" not in obj.PropertiesList:
                                obj.addProperty(
                                    "App::PropertyVectorList",
                                    "PolylinePoints",
                                    "Geometry",
                                    locked=True,
                                )
                            points = [p.Coordinates for p in ext.SweptArea.OuterCurve.Points]
                            points = [p + (0,) for p in points if len(p) < 3]
                            points = [FreeCAD.Vector(p).multiply(scaling) for p in points]
                            obj.PolylinePoints = points

                    # I profile
                    elif ext.SweptArea.is_a("IfcIShapeProfileDef"):
                        for p in [
                            "FilletRadius",
                            "FlangeEdgeRadius",
                            "FlangeSlope",
                            "FlangeThickness",
                            "OverallDepth",
                            "OverallWidth",
                            "WebThickness",
                        ]:
                            if hasattr(ext.SweptArea, p):
                                obj.addProperty("App::PropertyLength", p, "Geometry", locked=True)
                                value = getattr(ext.SweptArea, p)
                                if not value:
                                    value = 0
                                value = value * scaling
                                setattr(obj, p, value)
                            obj.addProperty(
                                "App::PropertyString", "ProfileName", "Geometry", locked=True
                            )
                            obj.ProfileName = ext.SweptArea.ProfileName

        # below is disabled for now... Don't know if it's useful to expose to the user
        elif False:  # rep.RepresentationIdentifier == "Axis":
            # Wall axis consisting on a single line
            if len(rep.Items) == 1:
                if rep.Items[0].is_a("IfcCompositeCurve"):
                    if len(rep.Items[0].Segments) == 1:
                        if rep.Items[0].Segments[0].is_a("IfcCompositeCurveSegment"):
                            if rep.Items[0].Segments[0].ParentCurve.is_a("IfcPolyline"):
                                pol = rep.Items[0].Segments[0].ParentCurve
                                if len(pol.Points) == 2:
                                    if "AxisStart" not in obj.PropertiesList:
                                        obj.addProperty(
                                            "App::PropertyPosition",
                                            "AxisStart",
                                            "Geometry",
                                            locked=True,
                                        )
                                    obj.AxisStart = FreeCAD.Vector(
                                        pol.Points[0].Coordinates
                                    ).multiply(scaling)
                                    if "AxisEnd" not in obj.PropertiesList:
                                        obj.addProperty(
                                            "App::PropertyPosition",
                                            "AxisEnd",
                                            "Geometry",
                                            locked=True,
                                        )
                                    obj.AxisEnd = FreeCAD.Vector(
                                        pol.Points[1].Coordinates
                                    ).multiply(scaling)


def set_attribute(ifcfile, element, prop, value):
    """Sets an attribute. Returns True if the attribute was changed"""

    if value != getattr(element, prop):
        ifc_tools.api_run(
            "attribute.edit_attributes",
            ifcfile,
            product=element,
            attributes={prop: value},
        )
        return True
    return False


def set_geom_property(obj, prop):
    """Updates the internal IFC file with the given value"""

    # TODO verify if values are different than the stored value before changing!

    element = ifc_tools.get_ifc_element(obj)
    if not ifc_tools.has_representation(element):
        return False
    ifcfile = ifc_tools.get_ifcfile(obj)
    scaling = ifcopenshell.util.unit.calculate_unit_scale(ifcfile)
    scaling = 0.001 / scaling
    changed = False

    if prop == "ExtrusionDepth":
        for rep in element.Representation.Representations:
            if rep.RepresentationIdentifier == "Body":
                if len(rep.Items) == 1:
                    if rep.Items[0].is_a("IfcExtrudedAreaSolid"):
                        elem = rep.Items[0]
                        value = getattr(obj, prop).Value * scaling
                        changed = set_attribute(ifcfile, elem, "Depth", value)

    elif prop == "ExtrusionDirection":
        for rep in element.Representation.Representations:
            if rep.RepresentationIdentifier == "Body":
                if len(rep.Items) == 1:
                    if rep.Items[0].is_a("IfcExtrudedAreaSolid"):
                        elem = rep.Items[0].ExtrudedDirection
                        value = tuple(getattr(obj, prop))
                        changed = set_attribute(ifcfile, elem, "DirectionRatios", value)

    elif prop == "RectangleLength":
        for rep in element.Representation.Representations:
            if rep.RepresentationIdentifier == "Body":
                if len(rep.Items) == 1:
                    if rep.Items[0].is_a("IfcExtrudedAreaSolid"):
                        elem = rep.Items[0].SweptArea
                        if elem.is_a("IfcRectangleProfileDef"):
                            value = getattr(obj, prop).Value * scaling
                            changed = set_attribute(ifcfile, elem, "XDim", value)

    elif prop == "RectangleWidth":
        for rep in element.Representation.Representations:
            if rep.RepresentationIdentifier == "Body":
                if len(rep.Items) == 1:
                    if rep.Items[0].is_a("IfcExtrudedAreaSolid"):
                        elem = rep.Items[0].SweptArea
                        if elem.is_a("IfcRectangleProfileDef"):
                            value = getattr(obj, prop).Value * scaling
                            changed = set_attribute(ifcfile, elem, "YDim", value)

    elif prop == "PolylinePoints":
        # TODO check values against existing
        for rep in element.Representation.Representations:
            if rep.RepresentationIdentifier == "Body":
                if len(rep.Items) == 1:
                    if rep.Items[0].is_a("IfcExtrudedAreaSolid"):
                        if rep.Items[0].SweptArea.is_a("IfcArbitraryClosedProfileDef"):
                            if rep.Items[0].SweptArea.OuterCurve.is_a("IfcPolyline"):
                                elem = rep.Items[0].SweptArea.OuterCurve
                                elem_points = elem.Points
                                psize = elem_points[0].Dim
                                points = getattr(obj, prop)
                                if len(points) > len(elem_points):
                                    for i in range(len(points) - len(elem_points)):
                                        p = ifc_tools.api_run(
                                            "root.create_entity",
                                            ifcfile,
                                            ifc_class="IfcCartesianPoint",
                                        )
                                        elem_points.append(p)
                                    elem.Points = elem_points
                                elif len(points) < len(elem_points):
                                    rest = []
                                    for i in range(len(elem_points) - len(points)):
                                        rest.append(elem_points.pop())
                                    elem.Points = elem_points
                                    for r in rest:
                                        ifc_tools.api_run("root.remove_product", ifcfile, product=r)
                                if len(points) == len(elem_points):
                                    for i in range(len(points)):
                                        v = FreeCAD.Vector(points[i]).multiply(scaling)
                                        coord = tuple(v)[:psize]
                                        ifc_tools.api_run(
                                            "attribute.edit_attributes",
                                            ifcfile,
                                            product=elem_points[i],
                                            attributes={"Coordinates": coord},
                                        )
                                    changed = True

    elif prop in [
        "FilletRadius",
        "FlangeEdgeRadius",
        "FlangeSlope",
        "FlangeThickness",
        "OverallDepth",
        "OverallWidth",
        "WebThickness",
    ]:
        for rep in element.Representation.Representations:
            if rep.RepresentationIdentifier == "Body":
                if len(rep.Items) == 1:
                    if rep.Items[0].SweptArea.is_a("IfcIShapeProfileDef"):
                        elem = rep.Items[0].SweptArea
                        value = getattr(obj, prop).Value * scaling
                        if value == 0 and getattr(elem, prop) is None:
                            value = None
                        changed = set_attribute(ifcfile, elem, prop, value)

    elif prop in ["ProfileName"]:
        for rep in element.Representation.Representations:
            if rep.RepresentationIdentifier == "Body":
                if len(rep.Items) == 1:
                    if rep.Items[0].SweptArea.is_a("IfcIShapeProfileDef"):
                        elem = rep.Items[0].SweptArea
                        value = obj.ProfileName
                        changed = set_attribute(ifcfile, elem, prop, value)

    if changed:
        FreeCAD.Console.PrintLog(
            "DEBUG: Changing prop"
            + obj.Label
            + ":"
            + str(prop)
            + "to"
            + str(getattr(obj, prop))
            + "\n"
        )
    return changed



import FreeCAD as App
import FreeCADGui
import Arch
import Draft
import Part
import Sketcher
from bimtests.TestArchBaseGui import TestArchBaseGui


class TestArchBuildingPartGui(TestArchBaseGui):

    def testBuildingPart(self):
        """Create a BuildingPart from a wall with a window and check its shape."""
        # Also regression test for:
        # https://github.com/FreeCAD/FreeCAD/issues/6178
        # operation = "Arch BuildingPart"
        # _msg("  Test '{}'".format(operation))
        # Most of the code below taken from testWindow function.
        line = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(3000, 0, 0))
        wall = Arch.makeWall(line)
        sk = App.ActiveDocument.addObject("Sketcher::SketchObject", "Sketch001")
        sk.Placement.Rotation = App.Rotation(App.Vector(1, 0, 0), 90)
        sk.addGeometry(Part.LineSegment(App.Vector(500, 800, 0), App.Vector(1500, 800, 0)))
        sk.addGeometry(Part.LineSegment(App.Vector(1500, 800, 0), App.Vector(1500, 2000, 0)))
        sk.addGeometry(Part.LineSegment(App.Vector(1500, 2000, 0), App.Vector(500, 2000, 0)))
        sk.addGeometry(Part.LineSegment(App.Vector(500, 2000, 0), App.Vector(500, 800, 0)))
        sk.addConstraint(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
        sk.addConstraint(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
        sk.addConstraint(Sketcher.Constraint("Coincident", 2, 2, 3, 1))
        sk.addConstraint(Sketcher.Constraint("Coincident", 3, 2, 0, 1))
        App.ActiveDocument.recompute()
        win = Arch.makeWindow(sk)
        Arch.removeComponents(win, host=wall)
        App.ActiveDocument.recompute()
        bp = Arch.makeBuildingPart()

        # Wall visibility works when standalone
        FreeCADGui.Selection.clearSelection()
        FreeCADGui.Selection.addSelection(self.doc_name, wall.Name)
        assert wall.Visibility
        FreeCADGui.runCommand("Std_ToggleVisibility", 0)
        App.ActiveDocument.recompute()
        assert not wall.Visibility
        FreeCADGui.runCommand("Std_ToggleVisibility", 0)
        assert wall.Visibility

        bp.Group = [wall]
        App.ActiveDocument.recompute()
        # Fails with OCC 7.5
        # self.assertTrue(len(bp.Shape.Faces) == 16, "'{}' failed".format(operation))

        # Wall visibility works when inside a BuildingPart
        FreeCADGui.runCommand("Std_ToggleVisibility", 0)
        App.ActiveDocument.recompute()
        assert not wall.Visibility
        FreeCADGui.runCommand("Std_ToggleVisibility", 0)
        assert wall.Visibility

        # Wall visibility works when BuildingPart Toggled
        FreeCADGui.Selection.clearSelection()
        FreeCADGui.Selection.addSelection(self.doc_name, bp.Name)
        FreeCADGui.runCommand("Std_ToggleVisibility", 0)
        assert not wall.Visibility
        FreeCADGui.runCommand("Std_ToggleVisibility", 0)
        assert wall.Visibility

        # Wall visibiity works inside group inside BuildingPart Toggled
        grp = App.ActiveDocument.addObject("App::DocumentObjectGroup", "Group")
        grp.Label = "Group"
        grp.Group = [wall]
        bp.Group = [grp]
        App.ActiveDocument.recompute()
        assert wall.Visibility
        FreeCADGui.runCommand("Std_ToggleVisibility", 0)
        App.ActiveDocument.recompute()
        assert not wall.Visibility
        FreeCADGui.runCommand("Std_ToggleVisibility", 0)
        App.ActiveDocument.recompute()
        assert wall.Visibility



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2018 Benjamin Alterauge (ageeye)                        *
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create PointArray objects.

The copies will be created at the points of a point object.
"""
## @package make_pointarray
# \ingroup draftmake
# \brief Provides functions to create PointArray objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftutils.gui_utils as gui_utils

from draftutils.messages import _err
from draftutils.translate import translate
from draftobjects.pointarray import PointArray

if App.GuiUp:
    from draftviewproviders.view_array import ViewProviderDraftArray
    from draftviewproviders.view_draftlink import ViewProviderDraftLink


def make_point_array(base_object, point_object, extra=None, use_link=True):
    """Make a Draft PointArray object.

    Create copies of a `base_object` at the points defined by
    a `point_object`.

    Parameters
    ----------
    base_object: Part::Feature or str
        Any of object that has a `Part::TopoShape` that can be duplicated.
        This means most 2D and 3D objects produced with any workbench.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.

    point_object: Part::Feature, Sketcher::SketchObject, Mesh::Feature,
                  Points::FeatureCustom or str
        The object must have vertices and/or points.

    extra: Base::Placement, Base::Vector3, or Base::Rotation, optional
        It defaults to `None`.
        If it is provided, it is an additional placement that is applied
        to each copy of the array.
        The input could be a full placement, just a vector indicating
        the additional translation, or just a rotation.

    Returns
    -------
    Part::FeaturePython
        A scripted object of type `'PointArray'`.
        Its `Shape` is a compound of the copies of the original object.

    None
        If there is a problem it will return `None`.
    """
    _name = "make_point_array"

    found, doc = utils.find_doc(App.activeDocument())
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return None

    found, base_object = utils.find_object(base_object, doc)
    if not found:
        _err(translate("draft", "Wrong input: base_object not in document."))
        return None

    found, point_object = utils.find_object(point_object, doc)
    if not found:
        _err(translate("draft", "Wrong input: point_object not in document."))
        return None

    if not (
        (hasattr(point_object, "Shape") and hasattr(point_object.Shape, "Vertexes"))
        or hasattr(point_object, "Mesh")
        or hasattr(point_object, "Points")
    ):
        _err(translate("draft", "Wrong input: object has the wrong type."))
        return None

    if not extra:
        extra = App.Placement()

    try:
        utils.type_check([(extra, (App.Placement, App.Vector, App.Rotation))], name=_name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a placement, a vector, or a rotation."))
        return None

    # Convert the vector or rotation to a full placement
    if isinstance(extra, App.Vector):
        extra = App.Placement(extra, App.Rotation())
    elif isinstance(extra, App.Rotation):
        extra = App.Placement(App.Vector(), extra)

    if use_link:
        # The PointArray class must be called in this special way
        # to make it a LinkArray
        new_obj = doc.addObject("Part::FeaturePython", "PointArray", PointArray(None), None, True)
    else:
        new_obj = doc.addObject("Part::FeaturePython", "PointArray")
        PointArray(new_obj)

    new_obj.Base = base_object
    new_obj.PointObject = point_object
    new_obj.ExtraPlacement = extra

    if App.GuiUp:
        if use_link:
            ViewProviderDraftLink(new_obj.ViewObject)
        else:
            new_obj.Proxy.execute(
                new_obj
            )  # Updates Count which is required for correct DiffuseColor.
            ViewProviderDraftArray(new_obj.ViewObject)
            gui_utils.format_object(new_obj, new_obj.Base)
            new_obj.ViewObject.Proxy.resetColors(new_obj.ViewObject)
        new_obj.Base.ViewObject.hide()
        gui_utils.select(new_obj)

    return new_obj


def makePointArray(base, ptlst):
    """Create PointArray. DEPRECATED. Use 'make_point_array'."""
    utils.use_instead("make_point_array")

    return make_point_array(base, ptlst)


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Fillet object."""
## @package fillet
# \ingroup draftobjects
# \brief Provides the object code for the Fillet object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from draftobjects.base import DraftObject
from draftutils import gui_utils


class Fillet(DraftObject):
    """Proxy class for the Fillet object."""

    def __init__(self, obj):
        super().__init__(obj, "Fillet")
        self._set_properties(obj)

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(obj, vp_module="view_fillet", vp_class="ViewProviderFillet")

    def _set_properties(self, obj):
        """Set the properties of objects if they don't exist."""
        if not hasattr(obj, "Start"):
            _tip = QT_TRANSLATE_NOOP("App::Property", "The start point of this line.")
            obj.addProperty("App::PropertyVectorDistance", "Start", "Draft", _tip, locked=True)
            obj.Start = App.Vector(0, 0, 0)

        if not hasattr(obj, "End"):
            _tip = QT_TRANSLATE_NOOP("App::Property", "The end point of this line.")
            obj.addProperty("App::PropertyVectorDistance", "End", "Draft", _tip, locked=True)
            obj.End = App.Vector(0, 0, 0)

        if not hasattr(obj, "Length"):
            _tip = QT_TRANSLATE_NOOP("App::Property", "The length of this line.")
            obj.addProperty("App::PropertyLength", "Length", "Draft", _tip, locked=True)
            obj.Length = 0

        if not hasattr(obj, "FilletRadius"):
            _tip = QT_TRANSLATE_NOOP("App::Property", "Radius to use to fillet the corner.")
            obj.addProperty("App::PropertyLength", "FilletRadius", "Draft", _tip, locked=True)
            obj.FilletRadius = 0

        # TODO: these two properties should link two straight lines
        # or edges so we can use them to build a fillet from them.
        # if not hasattr(obj, "Edge1"):
        #    _tip = "First line used as reference."
        #    obj.addProperty("App::PropertyLinkGlobal",
        #                    "Edge1",
        #                    "Draft",
        #                    _tip))

        # if not hasattr(obj, "Edge2"):
        #    _tip = "Second line used as reference."
        #    obj.addProperty("App::PropertyLinkGlobal",
        #                    "Edge2",
        #                    "Draft",
        #                    _tip))

        # Read only, change to 0 to make it editable.
        # The Fillet Radius should be made editable
        # when we are able to recalculate the arc of the fillet.
        obj.setEditorMode("Start", 1)
        obj.setEditorMode("End", 1)
        obj.setEditorMode("Length", 1)
        obj.setEditorMode("FilletRadius", 1)
        # obj.setEditorMode("Edge1", 1)
        # obj.setEditorMode("Edge2", 1)

    def execute(self, obj):
        """Run when the object is created or recomputed."""
        if hasattr(obj, "Length"):
            obj.Length = obj.Shape.Length
        if hasattr(obj, "Start"):
            obj.Start = obj.Shape.Vertexes[0].Point
        if hasattr(obj, "End"):
            obj.End = obj.Shape.Vertexes[-1].Point

    def _update_radius(self, obj, radius):
        # if (hasattr(obj, "Line1") and hasattr(obj, "Line2")
        #        and obj.Line1 and obj.Line2):
        # do the unimplemented work
        pass

    def onChanged(self, obj, prop):
        """Change the radius of fillet. NOT IMPLEMENTED.

        This should automatically recalculate the new fillet
        based on the new value of `FilletRadius`.
        """
        if prop in "FilletRadius":
            self._update_radius(obj, obj.FilletRadius)


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Arch Space"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchSpace
#  \ingroup ARCH
#  \brief The Space object and tools
#
#  This module provides tools to build Space objects.
#  Spaces define an open volume inside or outside a
#  building, ie. a room.

import re

import FreeCAD
import ArchComponent
import ArchCommands
import Draft

from draftutils import params

if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond

SpaceTypes = [
    "Undefined",
    "Exterior",
    "Exterior - Terrace",
    "Office",
    "Office - Enclosed",
    "Office - Open Plan",
    "Conference / Meeting / Multipurpose",
    "Classroom / Lecture / Training For Penitentiary",
    "Lobby",
    "Lobby - For Hotel",
    "Lobby - For Performing Arts Theater",
    "Lobby - For Motion Picture Theater",
    "Audience/Seating Area",
    "Audience/Seating Area - For Gymnasium",
    "Audience/Seating Area - For Exercise Center",
    "Audience/Seating Area - For Convention Center",
    "Audience/Seating Area - For Penitentiary",
    "Audience/Seating Area - For Religious Buildings",
    "Audience/Seating Area - For Sports Arena",
    "Audience/Seating Area - For Performing Arts Theater",
    "Audience/Seating Area - For Motion Picture Theater",
    "Audience/Seating Area - For Transportation",
    "Atrium",
    "Atrium - First Three Floors",
    "Atrium - Each Additional Floor",
    "Lounge / Recreation",
    "Lounge / Recreation - For Hospital",
    "Dining Area",
    "Dining Area - For Penitentiary",
    "Dining Area - For Hotel",
    "Dining Area - For Motel",
    "Dining Area - For Bar Lounge/Leisure Dining",
    "Dining Area - For Family Dining",
    "Food Preparation",
    "Laboratory",
    "Restrooms",
    "Dressing / Locker / Fitting",
    "Room",
    "Corridor / Transition",
    "Corridor / Transition - For Hospital",
    "Corridor / Transition - For Manufacturing Facility",
    "Stairs",
    "Active Storage",
    "Active Storage - For Hospital",
    "Inactive Storage",
    "Inactive Storage - For Museum",
    "Electrical / Mechanical",
    "Gymnasium / Exercise Center",
    "Gymnasium / Exercise Center - Playing Area",
    "Gymnasium / Exercise Center - Exercise Area",
    "Courthouse / Police Station / Penitentiary",
    "Courthouse / Police Station / Penitentiary - Courtroom",
    "Courthouse / Police Station / Penitentiary - Confinement Cells",
    "Courthouse / Police Station / Penitentiary - Judges' Chambers",
    "Fire Stations",
    "Fire Stations - Engine Room",
    "Fire Stations - Sleeping Quarters",
    "Post Office - Sorting Area",
    "Convention Center - Exhibit Space",
    "Library",
    "Library - Card File and Cataloging",
    "Library - Stacks",
    "Library - Reading Area",
    "Hospital",
    "Hospital - Emergency",
    "Hospital - Recovery",
    "Hospital - Nurses' Station",
    "Hospital - Exam / Treatment",
    "Hospital - Pharmacy",
    "Hospital - Patient Room",
    "Hospital - Operating Room",
    "Hospital - Nursery",
    "Hospital - Medical Supply",
    "Hospital - Physical Therapy",
    "Hospital - Radiology",
    "Hospital - Laundry-Washing",
    "Automotive - Service / Repair",
    "Manufacturing",
    "Manufacturing - Low Bay (< 7.5m Floor to Ceiling Height)",
    "Manufacturing - High Bay (> 7.5m Floor to Ceiling Height)",
    "Manufacturing - Detailed Manufacturing",
    "Manufacturing - Equipment Room",
    "Manufacturing - Control Room",
    "Hotel / Motel Guest Rooms",
    "Dormitory - Living Quarters",
    "Museum",
    "Museum - General Exhibition",
    "Museum - Restoration",
    "Bank / Office - Banking Activity Area",
    "Workshop",
    "Sales Area",
    "Religious Buildings",
    "Religious Buildings - Worship Pulpit, Choir",
    "Religious Buildings - Fellowship Hall",
    "Retail",
    "Retail - Sales Area",
    "Retail - Mall Concourse",
    "Sports Arena",
    "Sports Arena - Ring Sports Area",
    "Sports Arena - Court Sports Area",
    "Sports Arena - Indoor Playing Field Area",
    "Warehouse",
    "Warehouse - Fine Material Storage",
    "Warehouse - Medium / Bulky Material Storage",
    "Parking Garage - Garage Area",
    "Transportation",
    "Transportation - Airport / Concourse",
    "Transportation - Air / Train / Bus - Baggage Area",
    "Transportation - Terminal - Ticket Counter",
]

ConditioningTypes = [
    "Unconditioned",
    "Heated",
    "Cooled",
    "HeatedAndCooled",
    "Vented",
    "NaturallyVentedOnly",
]

AreaCalculationType = ["XY-plane projection", "At Center of Mass"]


class _Space(ArchComponent.Component):
    "A space object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "Space"
        self.setProperties(obj)
        obj.IfcType = "Space"
        obj.CompositionType = "ELEMENT"

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Boundaries" in pl:
            obj.addProperty(
                "App::PropertyLinkSubList",
                "Boundaries",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The objects that make the boundaries of this space object"
                ),
                locked=True,
            )
        if not "Area" in pl:
            obj.addProperty(
                "App::PropertyArea",
                "Area",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "Identical to Horizontal Area"),
                locked=True,
            )
        if not "FinishFloor" in pl:
            obj.addProperty(
                "App::PropertyString",
                "FinishFloor",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The finishing of the floor of this space"),
                locked=True,
            )
        if not "FinishWalls" in pl:
            obj.addProperty(
                "App::PropertyString",
                "FinishWalls",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The finishing of the walls of this space"),
                locked=True,
            )
        if not "FinishCeiling" in pl:
            obj.addProperty(
                "App::PropertyString",
                "FinishCeiling",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The finishing of the ceiling of this space"),
                locked=True,
            )
        if not "Group" in pl:
            obj.addProperty(
                "App::PropertyLinkList",
                "Group",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Objects that are included inside this space, such as furniture",
                ),
                locked=True,
            )
        if not "SpaceType" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "SpaceType",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The type of this space"),
                locked=True,
            )
            obj.SpaceType = SpaceTypes
        if not "FloorThickness" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "FloorThickness",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The thickness of the floor finish"),
                locked=True,
            )
        if not "NumberOfPeople" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "NumberOfPeople",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The number of people who typically occupy this space"
                ),
                locked=True,
            )
        if not "LightingPower" in pl:
            obj.addProperty(
                "App::PropertyFloat",
                "LightingPower",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The electric power needed to light this space in Watts"
                ),
                locked=True,
            )
        if not "EquipmentPower" in pl:
            obj.addProperty(
                "App::PropertyFloat",
                "EquipmentPower",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The electric power needed by the equipment of this space in Watts",
                ),
                locked=True,
            )
        if not "AutoPower" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "AutoPower",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If True, Equipment Power will be automatically filled by the equipment included in this space",
                ),
                locked=True,
            )
        if not "Conditioning" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "Conditioning",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The type of air conditioning of this space"),
                locked=True,
            )
            obj.Conditioning = ConditioningTypes
        if not "Internal" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "Internal",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Specifies if this space is internal or external"
                ),
                locked=True,
            )
            obj.Internal = True
        if not "AreaCalculationType" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "AreaCalculationType",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Defines the calculation type for the horizontal area and its perimeter length",
                ),
                locked=True,
            )
            obj.AreaCalculationType = AreaCalculationType

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

    def loads(self, state):

        self.Type = "Space"

    def execute(self, obj):

        if self.clone(obj):
            return

        # Space can do without Base.  Base validity is tested in getShape() code below.
        # Remarked out ensureBase() below
        # if not self.ensureBase(obj):
        #    return
        self.getShape(obj)

    def onChanged(self, obj, prop):

        if prop == "Group":
            if hasattr(obj, "EquipmentPower"):
                if obj.AutoPower:
                    p = 0
                    for o in Draft.getObjectsOfType(
                        Draft.get_group_contents(obj.Group, addgroups=True), "Equipment"
                    ):
                        if hasattr(o, "EquipmentPower"):
                            p += o.EquipmentPower
                    if p != obj.EquipmentPower:
                        obj.EquipmentPower = p
        elif prop == "Zone":
            if obj.Zone:
                if obj.Zone.ViewObject:
                    if hasattr(obj.Zone.ViewObject, "Proxy"):
                        if hasattr(obj.Zone.ViewObject.Proxy, "claimChildren"):
                            obj.Zone.ViewObject.Proxy.claimChildren()
        if hasattr(obj, "Area"):
            obj.setEditorMode("Area", 1)
        ArchComponent.Component.onChanged(self, obj, prop)

    def addSubobjects(self, obj, subobjects):
        "adds subobjects to this space"
        objs = obj.Boundaries
        for o in subobjects:
            if isinstance(o, tuple) or isinstance(o, list):
                if o[0].Name != obj.Name:
                    objs.append(tuple(o))
            else:
                for el in o.SubElementNames:
                    if "Face" in el:
                        if o.Object.Name != obj.Name:
                            objs.append((o.Object, el))
        obj.Boundaries = objs

    def removeSubobjects(self, obj, subobjects):
        "removes subobjects to this space"
        bounds = obj.Boundaries
        for o in subobjects:
            for b in bounds:
                if o.Name == b[0].Name:
                    bounds.remove(b)
                    break
        obj.Boundaries = bounds

    def addObject(self, obj, child):
        "Adds an object to this Space"

        if not child in obj.Group:
            g = obj.Group
            g.append(child)
            obj.Group = g

    def getShape(self, obj):
        "computes a shape from a base shape and/or boundary faces"
        import Part

        shape = None
        faces = []

        pl = obj.Placement

        # print("starting compute")

        # 1: if we have a base shape, we use it
        # Check if there is obj.Base and its validity to proceed
        if self.ensureBase(obj):
            if obj.Base.Shape.Solids:
                shape = obj.Base.Shape.copy()
                shape = shape.removeSplitter()

        # 2: if not, add all bounding boxes of considered objects and build a first shape
        if shape:
            # print("got shape from base object")
            bb = shape.BoundBox
        else:
            bb = None
            for b in obj.Boundaries:
                if hasattr(b[0], "Shape"):
                    if not bb:
                        bb = b[0].Shape.BoundBox
                    else:
                        bb.add(b[0].Shape.BoundBox)
            if not bb:
                # compute area even if we are not calculating the shape
                if obj.Shape and obj.Shape.Solids:
                    if hasattr(obj.Area, "Value"):
                        a = self.getArea(obj)
                        if obj.Area.Value != a:
                            obj.Area = a
                return
            shape = Part.makeBox(
                bb.XLength, bb.YLength, bb.ZLength, FreeCAD.Vector(bb.XMin, bb.YMin, bb.ZMin)
            )
            # print("created shape from boundbox")

        # 3: identifying boundary faces
        goodfaces = []
        for b in obj.Boundaries:
            if hasattr(b[0], "Shape"):
                for sub in b[1]:
                    if "Face" in sub:
                        fn = int(sub[4:]) - 1
                        faces.append(b[0].Shape.Faces[fn])
                        # print("adding face ",fn," of object ",b[0].Name)

        # print("total: ", len(faces), " faces")

        # 4: get cutvolumes from faces
        cutvolumes = []
        for f in faces:
            f = f.copy()
            f.reverse()
            cutface, cutvolume, invcutvolume = ArchCommands.getCutVolume(f, shape)
            if cutvolume:
                # print("generated 1 cutvolume")
                cutvolumes.append(cutvolume.copy())
                # Part.show(cutvolume)
        for v in cutvolumes:
            # print("cutting")
            shape = shape.cut(v)

        # 5: get the final shape
        if shape:
            if shape.Solids:
                # print("setting objects shape")
                shape = self.processSubShapes(obj, shape.Solids[0], pl)
                self.applyShape(obj, shape, pl)
                if hasattr(obj.HorizontalArea, "Value"):
                    if hasattr(obj, "AreaCalculationType"):
                        if obj.AreaCalculationType == "At Center of Mass":
                            a = self.getArea(obj)
                            obj.HorizontalArea = a
                    if hasattr(obj, "Area"):
                        obj.Area = obj.HorizontalArea

                return

        print("Arch: error computing space boundary for", obj.Label)

    def getArea(self, obj, notouch=False):
        "returns the horizontal area at the center of the space"

        self.face = self.getFootprint(obj)
        if self.face:
            if not notouch:
                if hasattr(obj, "PerimeterLength"):
                    if self.face.OuterWire.Length != obj.PerimeterLength.Value:
                        obj.PerimeterLength = self.face.OuterWire.Length
            return self.face.Area
        else:
            return 0

    def getFootprint(self, obj):
        "returns a face that represents the footprint of this space at the center of mass"

        import Part
        import DraftGeomUtils

        if not hasattr(obj.Shape, "CenterOfMass"):
            return None
        try:
            pl = Part.makePlane(1, 1)
            pl.translate(obj.Shape.CenterOfMass)
            sh = obj.Shape.copy()
            cutplane, v1, v2 = ArchCommands.getCutVolume(pl, sh)
            e = sh.section(cutplane)
            e = Part.__sortEdges__(e.Edges)
            w = Part.Wire(e)
            dv = FreeCAD.Vector(
                obj.Shape.CenterOfMass.x, obj.Shape.CenterOfMass.y, obj.Shape.BoundBox.ZMin
            )
            dv = dv.sub(obj.Shape.CenterOfMass)
            w.translate(dv)
            return Part.Face(w)
        except Part.OCCError:
            return None


class _ViewProviderSpace(ArchComponent.ViewProviderComponent):
    "A View Provider for Section Planes"

    def __init__(self, vobj):

        ArchComponent.ViewProviderComponent.__init__(self, vobj)
        self.setProperties(vobj)
        vobj.Transparency = params.get_param_arch("defaultSpaceTransparency")
        vobj.LineWidth = params.get_param_view("DefaultShapeLineWidth")
        vobj.LineColor = ArchCommands.getDefaultColor("Space")
        vobj.DrawStyle = ["Solid", "Dashed", "Dotted", "Dashdot"][
            params.get_param_arch("defaultSpaceStyle")
        ]

    def setProperties(self, vobj):

        pl = vobj.PropertiesList
        if not "Text" in pl:
            vobj.addProperty(
                "App::PropertyStringList",
                "Text",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The text to show. Use $area, $label, $longname, $description or any other property name preceded with $ (case insensitive), or $floor, $walls, $ceiling for finishes, to insert the respective data",
                ),
                locked=True,
            )
            vobj.Text = ["$label", "$area"]
        if not "FontName" in pl:
            vobj.addProperty(
                "App::PropertyFont",
                "FontName",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The name of the font"),
                locked=True,
            )
            vobj.FontName = params.get_param("textfont")
        if not "TextColor" in pl:
            vobj.addProperty(
                "App::PropertyColor",
                "TextColor",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The color of the area text"),
                locked=True,
            )
            vobj.TextColor = (0.0, 0.0, 0.0, 1.0)
        if not "FontSize" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "FontSize",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The size of the text font"),
                locked=True,
            )
            vobj.FontSize = params.get_param("textheight") * params.get_param(
                "DefaultAnnoScaleMultiplier"
            )
        if not "FirstLine" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "FirstLine",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The size of the first line of text"),
                locked=True,
            )
            vobj.FirstLine = params.get_param("textheight") * params.get_param(
                "DefaultAnnoScaleMultiplier"
            )
        if not "LineSpacing" in pl:
            vobj.addProperty(
                "App::PropertyFloat",
                "LineSpacing",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The space between the lines of text"),
                locked=True,
            )
            vobj.LineSpacing = 1.0
        if not "TextPosition" in pl:
            vobj.addProperty(
                "App::PropertyVectorDistance",
                "TextPosition",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The position of the text. Leave (0,0,0) for automatic position",
                ),
                locked=True,
            )
        if not "TextAlign" in pl:
            vobj.addProperty(
                "App::PropertyEnumeration",
                "TextAlign",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "The justification of the text"),
                locked=True,
            )
            vobj.TextAlign = ["Left", "Center", "Right"]
            vobj.TextAlign = "Center"
        if not "Decimals" in pl:
            vobj.addProperty(
                "App::PropertyInteger",
                "Decimals",
                "Space",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The number of decimals to use for calculated texts"
                ),
                locked=True,
            )
            vobj.Decimals = params.get_param("dimPrecision")
        if not "ShowUnit" in pl:
            vobj.addProperty(
                "App::PropertyBool",
                "ShowUnit",
                "Space",
                QT_TRANSLATE_NOOP("App::Property", "Show the unit suffix"),
                locked=True,
            )
            vobj.ShowUnit = params.get_param("showUnit")

    def onDocumentRestored(self, vobj):

        self.setProperties(vobj)

    def getIcon(self):

        import Arch_rc

        if hasattr(self, "Object"):
            if hasattr(self.Object, "CloneOf"):
                if self.Object.CloneOf:
                    return ":/icons/Arch_Space_Clone.svg"
        return ":/icons/Arch_Space_Tree.svg"

    def attach(self, vobj):

        ArchComponent.ViewProviderComponent.attach(self, vobj)
        from pivy import coin

        self.color = coin.SoBaseColor()
        self.font = coin.SoFont()
        self.text1 = coin.SoAsciiText()
        self.text1.string = " "
        self.text1.justification = coin.SoAsciiText.LEFT
        self.text2 = coin.SoAsciiText()
        self.text2.string = " "
        self.text2.justification = coin.SoAsciiText.LEFT
        self.coords = coin.SoTransform()
        self.header = coin.SoTransform()
        self.label = coin.SoSwitch()
        sep = coin.SoSeparator()
        self.label.whichChild = 0
        sep.addChild(self.coords)
        sep.addChild(self.color)
        sep.addChild(self.font)
        sep.addChild(self.text2)
        sep.addChild(self.header)
        sep.addChild(self.text1)
        self.label.addChild(sep)
        vobj.Annotation.addChild(self.label)
        self.onChanged(vobj, "TextColor")
        self.onChanged(vobj, "FontSize")
        self.onChanged(vobj, "FirstLine")
        self.onChanged(vobj, "LineSpacing")
        self.onChanged(vobj, "FontName")
        self.Object = vobj.Object
        # footprint mode
        self.fmat = coin.SoMaterial()
        self.fcoords = coin.SoCoordinate3()
        self.fset = coin.SoIndexedFaceSet()
        fhints = coin.SoShapeHints()
        fhints.vertexOrdering = fhints.COUNTERCLOCKWISE
        sep = coin.SoSeparator()
        sep.addChild(self.fmat)
        sep.addChild(self.fcoords)
        sep.addChild(fhints)
        sep.addChild(self.fset)
        vobj.RootNode.addChild(sep)

    def updateData(self, obj, prop):

        if prop in ["Shape", "Label", "Tag", "Area"]:
            self.onChanged(obj.ViewObject, "Text")
            self.onChanged(obj.ViewObject, "TextPosition")

    def getTextPosition(self, vobj):

        pos = FreeCAD.Vector()
        if hasattr(vobj, "TextPosition"):
            import DraftVecUtils

            if DraftVecUtils.isNull(vobj.TextPosition):
                try:
                    pos = vobj.Object.Shape.CenterOfMass
                    z = vobj.Object.Shape.BoundBox.ZMin
                    pos = FreeCAD.Vector(pos.x, pos.y, z)
                except (AttributeError, RuntimeError):
                    pos = FreeCAD.Vector()
            else:
                pos = vobj.Object.Placement.multVec(vobj.TextPosition)
        # placement's displacement will be already added by the coin node
        pos = vobj.Object.Placement.inverse().multVec(pos)
        return pos

    def onChanged(self, vobj, prop):

        if prop in ["Text", "Decimals", "ShowUnit"]:
            if hasattr(self, "text1") and hasattr(self, "text2") and hasattr(vobj, "Text"):
                self.text1.string.deleteValues(0)
                self.text2.string.deleteValues(0)
                text1 = []
                text2 = []
                first = True
                for t in vobj.Text:
                    if t:
                        t = t.replace("$label", vobj.Object.Label)
                        if hasattr(vobj.Object, "Area"):
                            from FreeCAD import Units

                            q = Units.Quantity(
                                vobj.Object.Area.Value, Units.Area
                            ).getUserPreferred()
                            qt = vobj.Object.Area.Value / q[1]
                            if hasattr(vobj, "Decimals"):
                                if vobj.Decimals == 0:
                                    qt = str(int(qt))
                                else:
                                    f = "%." + str(abs(vobj.Decimals)) + "f"
                                    qt = f % qt
                            else:
                                qt = str(qt)
                            if hasattr(vobj, "ShowUnit"):
                                if vobj.ShowUnit:
                                    qt = qt + q[2].replace("^2", "\xb2")  # square symbol
                            t = t.replace("$area", qt)
                        if hasattr(vobj.Object, "FinishFloor"):
                            t = t.replace("$floor", vobj.Object.FinishFloor)
                        if hasattr(vobj.Object, "FinishWalls"):
                            t = t.replace("$walls", vobj.Object.FinishWalls)
                        if hasattr(vobj.Object, "FinishCeiling"):
                            t = t.replace("$ceiling", vobj.Object.FinishCeiling)
                        # replace all other properties
                        props = vobj.Object.PropertiesList
                        lower_props = [p.lower() for p in props]
                        for rtag in re.findall(r"\$\w+", t):
                            lower_rtag = rtag[1:].lower()
                            if lower_rtag in lower_props:
                                prop = props[lower_props.index(lower_rtag)]
                                value = getattr(vobj.Object, prop, "")
                                if hasattr(value, "UserString"):
                                    value = value.UserString
                                elif hasattr(value, "Label"):
                                    value = value.Label
                                elif hasattr(value, "Name"):
                                    value = value.Name
                                t = t.replace(rtag, str(value))
                        if first:
                            text1.append(t)
                        else:
                            text2.append(t)
                    first = False
                if text1:
                    self.text1.string.setValues(text1)
                if text2:
                    self.text2.string.setValues(text2)

        elif prop == "FontName":
            if hasattr(self, "font") and hasattr(vobj, "FontName"):
                self.font.name = str(vobj.FontName)

        elif prop == "FontSize":
            if hasattr(self, "font") and hasattr(vobj, "FontSize"):
                self.font.size = vobj.FontSize.Value
                if hasattr(vobj, "FirstLine"):
                    scale = vobj.FirstLine.Value / vobj.FontSize.Value
                    self.header.scaleFactor.setValue([scale, scale, scale])
                    self.onChanged(vobj, "TextPosition")

        elif prop == "FirstLine":
            if hasattr(self, "header") and hasattr(vobj, "FontSize") and hasattr(vobj, "FirstLine"):
                scale = vobj.FirstLine.Value / vobj.FontSize.Value
                self.header.scaleFactor.setValue([scale, scale, scale])
                self.onChanged(vobj, "TextPosition")

        elif prop == "TextColor":
            if hasattr(self, "color") and hasattr(vobj, "TextColor"):
                c = vobj.TextColor
                self.color.rgb.setValue(c[0], c[1], c[2])

        elif prop == "TextPosition":
            if (
                hasattr(self, "coords")
                and hasattr(self, "header")
                and hasattr(vobj, "TextPosition")
                and hasattr(vobj, "FirstLine")
            ):
                pos = self.getTextPosition(vobj)
                self.coords.translation.setValue(
                    [pos.x, pos.y, pos.z + 0.01]
                )  # adding small z offset to separate from bottom face
                up = vobj.FirstLine.Value * vobj.LineSpacing
                self.header.translation.setValue([0, up, 0])

        elif prop == "LineSpacing":
            if hasattr(self, "text1") and hasattr(self, "text2") and hasattr(vobj, "LineSpacing"):
                self.text1.spacing = vobj.LineSpacing
                self.text2.spacing = vobj.LineSpacing
                self.onChanged(vobj, "TextPosition")

        elif prop == "TextAlign":
            if hasattr(self, "text1") and hasattr(self, "text2") and hasattr(vobj, "TextAlign"):
                from pivy import coin

                if vobj.TextAlign == "Center":
                    self.text1.justification = coin.SoAsciiText.CENTER
                    self.text2.justification = coin.SoAsciiText.CENTER
                elif vobj.TextAlign == "Right":
                    self.text1.justification = coin.SoAsciiText.RIGHT
                    self.text2.justification = coin.SoAsciiText.RIGHT
                else:
                    self.text1.justification = coin.SoAsciiText.LEFT
                    self.text2.justification = coin.SoAsciiText.LEFT

        elif prop == "Visibility":
            if vobj.Visibility:
                self.label.whichChild = 0
            else:
                self.label.whichChild = -1

        elif prop == "Transparency":
            if hasattr(vobj, "DisplayMode"):
                vobj.DisplayMode = "Wireframe" if vobj.Transparency == 100 else "Flat Lines"

    def setEdit(self, vobj, mode):
        if mode != 0:
            return None

        taskd = SpaceTaskPanel()
        taskd.obj = self.Object
        taskd.update()
        taskd.updateBoundaries()
        FreeCADGui.Control.showDialog(taskd)
        return True

    def getDisplayModes(self, vobj):

        modes = ArchComponent.ViewProviderComponent.getDisplayModes(self, vobj) + ["Footprint"]
        return modes

    def setDisplayMode(self, mode):

        self.fset.coordIndex.deleteValues(0)
        self.fcoords.point.deleteValues(0)
        if mode == "Footprint":
            if hasattr(self, "Object"):
                face = self.Object.Proxy.getFootprint(self.Object)
                if face:
                    verts = []
                    fdata = []
                    idx = 0
                    tri = face.tessellate(1)
                    for v in tri[0]:
                        verts.append([v.x, v.y, v.z])
                    for f in tri[1]:
                        fdata.extend([f[0] + idx, f[1] + idx, f[2] + idx, -1])
                    idx += len(tri[0])
                    self.fcoords.point.setValues(verts)
                    self.fset.coordIndex.setValues(0, len(fdata), fdata)
            return "Points"
        return ArchComponent.ViewProviderComponent.setDisplayMode(self, mode)


class SpaceTaskPanel(ArchComponent.ComponentTaskPanel):
    "A modified version of the Arch component task panel"

    def __init__(self):

        ArchComponent.ComponentTaskPanel.__init__(self)
        self.editButton = QtGui.QPushButton(self.form)
        self.editButton.setObjectName("editButton")
        self.editButton.setIcon(QtGui.QIcon(":/icons/Draft_Edit.svg"))
        self.grid.addWidget(self.editButton, 4, 0, 1, 2)
        self.editButton.setText(QtGui.QApplication.translate("Arch", "Set text position", None))
        QtCore.QObject.connect(self.editButton, QtCore.SIGNAL("clicked()"), self.setTextPos)
        boundLabel = QtGui.QLabel(self.form)
        self.grid.addWidget(boundLabel, 5, 0, 1, 2)
        boundLabel.setText(QtGui.QApplication.translate("Arch", "Space boundaries", None))
        self.boundList = QtGui.QListWidget(self.form)
        self.grid.addWidget(self.boundList, 6, 0, 1, 2)
        self.addCompButton = QtGui.QPushButton(self.form)
        self.addCompButton.setObjectName("addCompButton")
        self.addCompButton.setIcon(QtGui.QIcon(":/icons/Arch_Add.svg"))
        self.grid.addWidget(self.addCompButton, 7, 0, 1, 1)
        self.addCompButton.setText(QtGui.QApplication.translate("Arch", "Add", None))
        QtCore.QObject.connect(self.addCompButton, QtCore.SIGNAL("clicked()"), self.addBoundary)
        self.delCompButton = QtGui.QPushButton(self.form)
        self.delCompButton.setObjectName("delCompButton")
        self.delCompButton.setIcon(QtGui.QIcon(":/icons/Arch_Remove.svg"))
        self.grid.addWidget(self.delCompButton, 7, 1, 1, 1)
        self.delCompButton.setText(QtGui.QApplication.translate("Arch", "Remove", None))
        QtCore.QObject.connect(self.delCompButton, QtCore.SIGNAL("clicked()"), self.delBoundary)

    def updateBoundaries(self):

        self.boundList.clear()
        if self.obj:
            for b in self.obj.Boundaries:
                s = b[0].Label
                for n in b[1]:
                    s += ", " + n
                it = QtGui.QListWidgetItem(s)
                it.setToolTip(b[0].Name)
                self.boundList.addItem(it)

    def setTextPos(self):

        FreeCADGui.runCommand("Draft_Edit")

    def addBoundary(self):

        if self.obj:
            if FreeCADGui.Selection.getSelectionEx():
                self.obj.Proxy.addSubobjects(self.obj, FreeCADGui.Selection.getSelectionEx())
                self.updateBoundaries()

    def delBoundary(self):

        if self.boundList.currentRow() >= 0:
            it = self.boundList.item(self.boundList.currentRow())
            if it and self.obj:
                on = it.toolTip()
                bounds = self.obj.Boundaries
                for b in bounds:
                    if b[0].Name == on:
                        bounds.remove(b)
                        break
                self.obj.Boundaries = bounds
                self.updateBoundaries()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2001, 2002 Juergen Riegel <juergen.riegel@web.de>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

# FreeCAD init script of the part module

# FreeCAD.addImportType("CAD formats (*.igs *.iges *.step *.stp *.brep *.brp)","Part")
# FreeCAD.addExportType("CAD formats (*.igs *.iges *.step *.stp *.brep *.brp)","Part")
FreeCAD.addImportType("BREP format (*.brep *.BREP *.brp *.BRP)", "Part")
FreeCAD.addExportType("BREP format (*.brep *.brp)", "Part")
FreeCAD.addImportType("IGES format (*.iges *.IGES *.igs *.IGS)", "Part")
FreeCAD.addExportType("IGES format (*.iges *.igs)", "Part")
FreeCAD.addImportType("STEP with colors (*.step *.STEP *.stp *.STP)", "Import")
FreeCAD.addExportType("STEP with colors (*.step *.stp)", "Import")

FreeCAD.__unit_test__ += ["TestPartApp"]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for all annotation type objects.

This is inherited and used by many viewproviders that show dimensions
and text created on screen through Coin (pivy).
- DimensionBase
- LinearDimension
- AngularDimension
- Label
- Text

Its edit mode launches the `Draft_Edit` command.
"""
## @package view_draft_annotation
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for all annotation type objects.

## \addtogroup draftviewproviders
# @{
import json
from PySide.QtCore import QT_TRANSLATE_NOOP

import PySide.QtCore as QtCore
import PySide.QtGui as QtGui

import FreeCAD as App
import FreeCADGui as Gui
from draftutils import params
from draftutils import utils
from draftutils.translate import translate


class ViewProviderDraftAnnotation(object):
    """The base class for Draft Annotation viewproviders.

    This class is intended to be inherited by more specialized viewproviders,
    for example, `Text`, `LinearDimension`, `RadialDimension`, and `Label`.
    """

    def __init__(self, vobj):
        self.set_properties(vobj)
        self.Object = vobj.Object
        vobj.Proxy = self

    def set_properties(self, vobj):
        """Set the properties only if they don't already exist."""
        properties = vobj.PropertiesList
        self.set_annotation_properties(vobj, properties)
        self.set_text_properties(vobj, properties)
        self.set_units_properties(vobj, properties)
        self.set_graphics_properties(vobj, properties)

    def set_annotation_properties(self, vobj, properties):
        """Set annotation properties only if they don't already exist."""
        if "ScaleMultiplier" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "General scaling factor that affects "
                "the annotation consistently\n"
                "because it scales the text, "
                "and the line decorations, if any,\n"
                "in the same proportion.",
            )
            vobj.addProperty(
                "App::PropertyFloat", "ScaleMultiplier", "Annotation", _tip, locked=True
            )
            vobj.ScaleMultiplier = params.get_param("DefaultAnnoScaleMultiplier")

        if "AnnotationStyle" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Annotation style to apply "
                "to this object.\n"
                "When using a saved style "
                "some of the view properties "
                "will become read-only;\n"
                "they will only be editable by changing "
                "the style through "
                "the 'Annotation style editor' tool.",
            )
            vobj.addProperty(
                "App::PropertyEnumeration", "AnnotationStyle", "Annotation", _tip, locked=True
            )
            styles = []
            for key in vobj.Object.Document.Meta.keys():
                if key.startswith("Draft_Style_"):
                    styles.append(key[12:])

            vobj.AnnotationStyle = [""] + styles

    def set_text_properties(self, vobj, properties):
        """Set text properties only if they don't already exist."""
        if "FontName" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Font name")
            vobj.addProperty("App::PropertyFont", "FontName", "Text", _tip, locked=True)
            vobj.FontName = params.get_param("textfont")

        if "FontSize" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Font size")
            vobj.addProperty("App::PropertyLength", "FontSize", "Text", _tip, locked=True)
            vobj.FontSize = params.get_param("textheight")

        if "TextColor" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Text color")
            vobj.addProperty("App::PropertyColor", "TextColor", "Text", _tip, locked=True)
            vobj.TextColor = params.get_param("DefaultTextColor") | 0x000000FF

    def set_units_properties(self, vobj, properties):
        return

    def set_graphics_properties(self, vobj, properties):
        """Set graphics properties only if they don't already exist."""
        if vobj.Object.Proxy.Type == "Text":
            return

        if "ArrowSizeStart" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Arrow size")
            vobj.addProperty("App::PropertyLength", "ArrowSizeStart", "Graphics", _tip, locked=True)
            vobj.ArrowSizeStart = params.get_param("arrowsizestart")

        if "ArrowTypeStart" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Arrow type")
            vobj.addProperty(
                "App::PropertyEnumeration", "ArrowTypeStart", "Graphics", _tip, locked=True
            )
            vobj.ArrowTypeStart = utils.ARROW_TYPES
            vobj.ArrowTypeStart = "None"

        if vobj.Object.Proxy.Type != "Label":

            if "ArrowSizeEnd" not in properties:
                _tip = QT_TRANSLATE_NOOP("App::Property", "Arrow size")
                vobj.addProperty(
                    "App::PropertyLength", "ArrowSizeEnd", "Graphics", _tip, locked=True
                )
                vobj.ArrowSizeEnd = params.get_param("arrowsizeend")

            if "ArrowTypeEnd" not in properties:
                _tip = QT_TRANSLATE_NOOP("App::Property", "Arrow type")
                vobj.addProperty(
                    "App::PropertyEnumeration", "ArrowTypeEnd", "Graphics", _tip, locked=True
                )
                vobj.ArrowTypeEnd = utils.ARROW_TYPES
                vobj.ArrowTypeEnd = "None"

        if "LineWidth" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Line width")
            vobj.addProperty("App::PropertyFloat", "LineWidth", "Graphics", _tip, locked=True)
            vobj.LineWidth = params.get_param("DefaultAnnoLineWidth")

        if "LineColor" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Line color")
            vobj.addProperty("App::PropertyColor", "LineColor", "Graphics", _tip, locked=True)
            vobj.LineColor = params.get_param("DefaultAnnoLineColor") | 0x000000FF

    def dumps(self):
        """Return a tuple of objects to save or None."""
        return None

    def loads(self, state):
        """Set the internal properties from the restored state."""
        return None

    def attach(self, vobj):
        """Set up the scene sub-graph of the view provider."""
        self.Object = vobj.Object

    def updateData(self, obj, prop):
        """Execute when a property from the Proxy class is changed."""
        return

    def getDisplayModes(self, vobj):
        """Return the display modes that this viewprovider supports."""
        return ["World", "Screen"]

    def getDefaultDisplayMode(self):
        """Return the default display mode."""
        return ["World", "Screen"][params.get_param("DefaultAnnoDisplayMode")]

    def setDisplayMode(self, mode):
        """Return the saved display mode."""
        return mode

    def onChanged(self, vobj, prop):
        """Execute when a view property is changed."""
        properties = vobj.PropertiesList
        meta = vobj.Object.Document.Meta

        if prop == "AnnotationStyle" and "AnnotationStyle" in properties:
            if not vobj.AnnotationStyle or vobj.AnnotationStyle == "":
                # unset style
                for visprop in utils.get_default_annotation_style().keys():
                    if visprop in properties:
                        # make property writable
                        vobj.setPropertyStatus(visprop, "-ReadOnly")
            else:
                # set style
                for key, value in meta.items():
                    if key.startswith("Draft_Style_") and key[12:] == vobj.AnnotationStyle:
                        style = utils.repair_annotation_style(json.loads(value))
                        for visprop in style.keys():
                            if visprop in properties:
                                # make property read-only
                                vobj.setPropertyStatus(visprop, "ReadOnly")
                                value = style[visprop]
                                try:
                                    if vobj.getTypeIdOfProperty(visprop) == "App::PropertyColor":
                                        value = value | 0x000000FF
                                    setattr(vobj, visprop, value)
                                except TypeError:
                                    pass
                        break

    def execute(self, vobj):
        """Execute when the object is created or recomputed."""
        return

    def setEdit(self, vobj, mode):
        """Execute the code when entering the specific edit mode.

        See view_base.py.
        """
        if mode != 0:
            return None

        if utils.get_type(vobj.Object) == "AngularDimension":
            return False  # Required, else edit mode is entered.

        if not "Draft_Edit" in Gui.listCommands():
            self.wb_before_edit = Gui.activeWorkbench()
            Gui.activateWorkbench("DraftWorkbench")
        Gui.runCommand("Draft_Edit")
        return True

    def unsetEdit(self, vobj, mode):
        """Execute the code when leaving the specific edit mode.

        See view_base.py.
        """
        if mode != 0:
            return None

        if hasattr(App, "activeDraftCommand") and App.activeDraftCommand:
            App.activeDraftCommand.finish()
        Gui.Control.closeDialog()
        if hasattr(self, "wb_before_edit"):
            Gui.activateWorkbench(self.wb_before_edit.name())
            delattr(self, "wb_before_edit")
        return True

    def doubleClicked(self, vobj):
        self.edit()

    def setupContextMenu(self, vobj, menu):
        if utils.get_type(self.Object) in ("AngularDimension", "Label"):
            return

        action_edit = QtGui.QAction(translate("draft", "Edit"), menu)
        QtCore.QObject.connect(action_edit, QtCore.SIGNAL("triggered()"), self.edit)
        menu.addAction(action_edit)

    def edit(self):
        Gui.ActiveDocument.setEdit(self.Object, 0)

    def getIcon(self):
        """Return the path to the icon used by the view provider."""
        return ":/icons/Draft_Text.svg"


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM Panel-related Arch_"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


#           Description                 l    w    t

Presets = [
    None,
    ["Plywood 12mm, 1220 x 2440", 1220, 2440, 12],
    ["Plywood 15mm, 1220 x 2440", 1220, 2440, 15],
    ["Plywood 18mm, 1220 x 2440", 1220, 2440, 18],
    ["Plywood 25mm, 1220 x 2440", 1220, 2440, 25],
    ["MDF 3mm, 900 x 600", 900, 600, 3],
    ["MDF 6mm, 900 x 600", 900, 600, 6],
    ["OSB 18mm, 1220 x 2440", 1220, 2440, 18],
]


class Arch_Panel:
    "the Arch Panel Arch_ definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Panel",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Panel", "Panel"),
            "Accel": "P, A",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Panel",
                "Creates a panel object from scratch or from a selected object (sketch, wire, face or solid)",
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        import WorkingPlane
        import Draft
        import draftguitools.gui_trackers as DraftTrackers
        from draftutils import params

        self.doc = FreeCAD.ActiveDocument
        self.Length = params.get_param_arch("PanelLength")
        self.Width = params.get_param_arch("PanelWidth")
        self.Thickness = params.get_param_arch("PanelThickness")
        self.Profile = None
        self.featureName = "Panel"
        self.rotated = False
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            if len(sel) == 1:
                if Draft.getType(sel[0]) == "Panel":
                    return
            self.doc.openTransaction(str(translate("Arch", "Create Panel")))
            FreeCADGui.addModule("Arch")
            FreeCADGui.addModule("Draft")
            for obj in sel:
                FreeCADGui.doCommand(
                    "obj = Arch.makePanel(FreeCAD.ActiveDocument."
                    + obj.Name
                    + ",thickness="
                    + str(self.Thickness)
                    + ")"
                )
                FreeCADGui.doCommand("Draft.autogroup(obj)")
            self.doc.commitTransaction()
            self.doc.recompute()
            return

        # interactive mode

        FreeCAD.activeDraftCommand = self  # register as a Draft command for auto grid on/off
        WorkingPlane.get_working_plane()
        self.points = []
        self.tracker = DraftTrackers.boxTracker()
        self.tracker.width(self.Width)
        self.tracker.height(self.Thickness)
        self.tracker.length(self.Length)
        self.tracker.on()
        FreeCADGui.Snapper.getPoint(
            callback=self.getPoint, movecallback=self.update, extradlg=self.taskbox()
        )
        FreeCADGui.draftToolBar.continueCmd.show()

    def getPoint(self, point=None, obj=None):
        "this function is called by the snapper when it has a 3D point"

        import DraftVecUtils

        FreeCAD.activeDraftCommand = None
        FreeCADGui.Snapper.off()
        self.tracker.finalize()
        if point is None:
            return
        self.doc.openTransaction(translate("Arch", "Create Panel"))
        FreeCADGui.addModule("Arch")
        if self.Profile:
            pr = Presets[self.Profile]
            FreeCADGui.doCommand(
                "p = Arch.makeProfile("
                + str(pr[2])
                + ","
                + str(pr[3])
                + ","
                + str(pr[4])
                + ","
                + str(pr[5])
                + ")"
            )
            FreeCADGui.doCommand("s = Arch.makePanel(p,thickness=" + str(self.Thickness) + ")")
            # FreeCADGui.doCommand('s.Placement.Rotation = FreeCAD.Rotation(-0.5,0.5,-0.5,0.5)')
        else:
            FreeCADGui.doCommand(
                "s = Arch.makePanel(length="
                + str(self.Length)
                + ",width="
                + str(self.Width)
                + ",thickness="
                + str(self.Thickness)
                + ")"
            )
        FreeCADGui.doCommand("s.Placement.Base = " + DraftVecUtils.toString(point))
        if self.rotated:
            FreeCADGui.doCommand(
                "s.Placement.Rotation = FreeCAD.Rotation(FreeCAD.Vector(1.00,0.00,0.00),90.00)"
            )
        self.doc.commitTransaction()
        self.doc.recompute()
        from PySide import QtCore

        QtCore.QTimer.singleShot(100, self.check_continueMode)

    def check_continueMode(self):
        "checks if continueMode is true and restarts Panel"

        if FreeCADGui.draftToolBar.continueMode:
            self.Activated()

    def taskbox(self):
        "sets up a taskbox widget"

        from draftutils import params
        from PySide import QtCore, QtGui

        w = QtGui.QWidget()
        ui = FreeCADGui.UiLoader()
        w.setWindowTitle(translate("Arch", "Panel options"))
        grid = QtGui.QGridLayout(w)

        # presets box
        labelp = QtGui.QLabel(translate("Arch", "Preset"))
        valuep = QtGui.QComboBox()
        fpresets = [" "]
        for p in Presets[1:]:
            fpresets.append(translate("Arch", p[0]))
        valuep.addItems(fpresets)
        grid.addWidget(labelp, 0, 0, 1, 1)
        grid.addWidget(valuep, 0, 1, 1, 1)

        # length
        label1 = QtGui.QLabel(translate("Arch", "Length"))
        self.vLength = ui.createWidget("Gui::InputField")
        self.vLength.setText(FreeCAD.Units.Quantity(self.Length, FreeCAD.Units.Length).UserString)
        grid.addWidget(label1, 1, 0, 1, 1)
        grid.addWidget(self.vLength, 1, 1, 1, 1)

        # width
        label2 = QtGui.QLabel(translate("Arch", "Width"))
        self.vWidth = ui.createWidget("Gui::InputField")
        self.vWidth.setText(FreeCAD.Units.Quantity(self.Width, FreeCAD.Units.Length).UserString)
        grid.addWidget(label2, 2, 0, 1, 1)
        grid.addWidget(self.vWidth, 2, 1, 1, 1)

        # height
        label3 = QtGui.QLabel(translate("Arch", "Thickness"))
        self.vHeight = ui.createWidget("Gui::InputField")
        self.vHeight.setText(
            FreeCAD.Units.Quantity(self.Thickness, FreeCAD.Units.Length).UserString
        )
        grid.addWidget(label3, 3, 0, 1, 1)
        grid.addWidget(self.vHeight, 3, 1, 1, 1)

        # horizontal button
        value4 = QtGui.QPushButton(translate("Arch", "Rotate"))
        grid.addWidget(value4, 4, 0, 1, 2)

        valuep.currentIndexChanged.connect(self.setPreset)
        self.vLength.valueChanged.connect(self.setLength)
        self.vWidth.valueChanged.connect(self.setWidth)
        self.vHeight.valueChanged.connect(self.setThickness)
        value4.pressed.connect(self.rotate)
        return w

    def update(self, point, info):
        "this function is called by the Snapper when the mouse is moved"

        if FreeCADGui.Control.activeDialog():
            self.tracker.pos(point)
            if self.rotated:
                self.tracker.width(self.Thickness)
                self.tracker.height(self.Width)
                self.tracker.length(self.Length)
            else:
                self.tracker.width(self.Width)
                self.tracker.height(self.Thickness)
                self.tracker.length(self.Length)

    def setWidth(self, d):

        from draftutils import params

        self.Width = d.Value
        params.set_param_arch("PanelWidth", d)

    def setThickness(self, d):

        from draftutils import params

        self.Thickness = d.Value
        params.set_param_arch("PanelThickness", d)

    def setLength(self, d):

        from draftutils import params

        self.Length = d.Value
        params.set_param_arch("PanelLength", d)

    def setPreset(self, i):

        if i > 0:
            self.vLength.setText(
                FreeCAD.Units.Quantity(float(Presets[i][1]), FreeCAD.Units.Length).UserString
            )
            self.vWidth.setText(
                FreeCAD.Units.Quantity(float(Presets[i][2]), FreeCAD.Units.Length).UserString
            )
            self.vHeight.setText(
                FreeCAD.Units.Quantity(float(Presets[i][3]), FreeCAD.Units.Length).UserString
            )

    def rotate(self):

        self.rotated = not self.rotated


class Arch_PanelCut:
    "the Arch Panel Cut command definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Panel_Cut",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Panel_Cut", "Panel Cut"),
            "Accel": "P, C",
            "ToolTip": QT_TRANSLATE_NOOP("Arch_Panel_Cut", "Creates 2D views of selected panels"),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        import Draft

        if FreeCADGui.Selection.getSelection():
            FreeCAD.ActiveDocument.openTransaction(str(translate("Arch", "Create Panel Cut")))
            FreeCADGui.addModule("Arch")
            for obj in FreeCADGui.Selection.getSelection():
                if Draft.getType(obj) == "Panel":
                    FreeCADGui.doCommand(
                        "Arch.makePanelCut(FreeCAD.ActiveDocument." + obj.Name + ")"
                    )
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()


class Arch_PanelSheet:
    "the Arch Panel Sheet definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Panel_Sheet",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Panel_Sheet", "Panel Sheet"),
            "Accel": "P, S",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Panel_Sheet", "Creates a 2D sheet which can contain panel cuts"
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        FreeCAD.ActiveDocument.openTransaction(str(translate("Arch", "Create Panel Sheet")))
        FreeCADGui.addModule("Arch")
        if FreeCADGui.Selection.getSelection():
            l = "["
            for obj in FreeCADGui.Selection.getSelection():
                l += "FreeCAD.ActiveDocument." + obj.Name + ","
            l += "]"
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()
            FreeCADGui.doCommand("__objs__ = " + l)
            FreeCADGui.doCommand("Arch.makePanelSheet(__objs__)")
            FreeCADGui.doCommand("del __objs__")
        else:
            FreeCADGui.doCommand("Arch.makePanelSheet()")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


class Arch_Nest:
    "the Arch Panel definition"

    def GetResources(self):

        return {
            "Pixmap": "Arch_Nest",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Nest", "Nest"),
            "Accel": "N, E",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Nest", "Nests a series of selected shapes in a container"
            ),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        FreeCADGui.Control.closeDialog()
        FreeCADGui.Control.showDialog(NestTaskPanel())


class NestTaskPanel:
    """The TaskPanel for Arch Nest"""

    def __init__(self, obj=None):

        import ArchNesting

        self.form = FreeCADGui.PySideUic.loadUi(":/ui/ArchNest.ui")
        self.form.progressBar.hide()
        self.form.ButtonPreview.setEnabled(False)
        self.form.ButtonStop.setEnabled(False)
        self.form.ButtonContainer.pressed.connect(self.getContainer)
        self.form.ButtonShapes.pressed.connect(self.getShapes)
        self.form.ButtonRemove.pressed.connect(self.removeShapes)
        self.form.ButtonStart.pressed.connect(self.start)
        self.form.ButtonStop.pressed.connect(self.stop)
        self.form.ButtonPreview.pressed.connect(self.preview)
        self.shapes = []
        self.container = None
        self.nester = None
        self.temps = []

    def reject(self):

        self.stop()
        self.clearTemps()
        return True

    def accept(self):

        self.stop()
        self.clearTemps()
        if self.nester:
            FreeCAD.ActiveDocument.openTransaction("Nesting")
            self.nester.apply()
            FreeCAD.ActiveDocument.commitTransaction()
        return True

    def clearTemps(self):

        for t in self.temps:
            if FreeCAD.ActiveDocument.getObject(t.Name):
                FreeCAD.ActiveDocument.removeObject(t.Name)
        self.temps = []

    def getContainer(self):

        import Draft

        s = FreeCADGui.Selection.getSelection()
        if len(s) == 1:
            if hasattr(s[0], "Shape"):
                if len(s[0].Shape.Faces) == 1:
                    if not (s[0] in self.shapes):
                        self.form.Container.clear()
                        self.addObject(s[0], self.form.Container)
                        self.container = s[0]
                else:
                    FreeCAD.Console.PrintError(translate("Arch", "This object has no face"))
                if Draft.getType(s[0]) == "PanelSheet":
                    if hasattr(s[0], "Rotations"):
                        if s[0].Rotations:
                            self.form.Rotations.setText(str(s[0].Rotations))

    def getShapes(self):

        s = FreeCADGui.Selection.getSelection()
        for o in s:
            if hasattr(o, "Shape"):
                if not o in self.shapes:
                    if o != self.container:
                        self.addObject(o, self.form.Shapes)
                        self.shapes.append(o)

    def addObject(self, obj, form):

        from PySide import QtGui

        i = QtGui.QListWidgetItem()
        i.setText(obj.Label)
        i.setToolTip(obj.Name)
        if hasattr(obj.ViewObject, "Proxy"):
            if hasattr(obj.ViewObject.Proxy, "getIcon"):
                i.setIcon(QtGui.QIcon(obj.ViewObject.Proxy.getIcon()))
        elif hasattr(obj.ViewObject, "Icon"):
            i.setIcon(QtGui.QIcon(obj.ViewObject.Icon))
        else:
            i.setIcon(QtGui.QIcon(":/icons/Part_3D_object.svg"))
        form.addItem(i)

    def removeShapes(self):

        for i in self.form.Shapes.selectedItems():
            o = FreeCAD.ActiveDocument.getObject(i.toolTip())
            if o:
                if o in self.shapes:
                    self.shapes.remove(o)
                    self.form.Shapes.takeItem(self.form.Shapes.row(i))

    def setCounter(self, value):

        self.form.progressBar.setValue(value)

    def start(self):

        from PySide import QtGui
        import ArchNesting

        self.clearTemps()
        self.form.progressBar.setFormat("pass 1: %p%")
        self.form.progressBar.setValue(0)
        self.form.progressBar.show()
        tolerance = self.form.Tolerance.value()
        discretize = self.form.Subdivisions.value()
        rotations = [float(x) for x in self.form.Rotations.text().split(",")]
        ArchNesting.TOLERANCE = tolerance
        ArchNesting.DISCRETIZE = discretize
        ArchNesting.ROTATIONS = rotations
        self.nester = ArchNesting.Nester()
        self.nester.addContainer(self.container)
        self.nester.addObjects(self.shapes)
        self.nester.setCounter = self.setCounter
        self.form.ButtonStop.setEnabled(True)
        self.form.ButtonStart.setEnabled(False)
        self.form.ButtonPreview.setEnabled(False)
        QtGui.QApplication.processEvents()
        result = self.nester.run()
        self.form.progressBar.hide()
        self.form.ButtonStart.setEnabled(True)
        self.form.ButtonStop.setEnabled(False)
        if result:
            self.form.ButtonPreview.setEnabled(True)

    def stop(self):

        if self.nester:
            self.nester.stop()
        self.form.ButtonStart.setEnabled(True)
        self.form.ButtonStop.setEnabled(False)
        self.form.ButtonPreview.setEnabled(False)
        self.form.progressBar.hide()

    def preview(self):

        self.clearTemps()
        if self.nester:
            t = self.nester.show()
            if t:
                self.temps.extend(t)


class Arch_PanelGroup:

    def GetCommands(self):
        return tuple(["Arch_Panel", "Arch_Panel_Cut", "Arch_Panel_Sheet", "Arch_Nest"])

    def GetResources(self):
        return {
            "MenuText": QT_TRANSLATE_NOOP("Arch_PanelTools", "Panel Tools"),
            "ToolTip": QT_TRANSLATE_NOOP("Arch_PanelTools", "Panel tools"),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v


FreeCADGui.addCommand("Arch_Panel", Arch_Panel())
FreeCADGui.addCommand("Arch_Panel_Cut", Arch_PanelCut())
FreeCADGui.addCommand("Arch_Panel_Sheet", Arch_PanelSheet())
FreeCADGui.addCommand("Arch_Nest", Arch_Nest())
FreeCADGui.addCommand("Arch_PanelTools", Arch_PanelGroup())



# SPDX-License-Identifier: LGPL-2.1-or-later

#***************************************************************************
#*   Copyright (c) 2001,2002 Juergen Riegel <juergen.riegel@web.de>        *
#*                                                                         *
#*   This file is part of the FreeCAD CAx development system.              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   FreeCAD is distributed in the hope that it will be useful,            *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Lesser General Public License for more details.                   *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with FreeCAD; if not, write to the Free Software        *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************/

# FreeCAD init script of the OpenSCAD module

import os
import FreeCAD

FreeCAD.addImportType("OpenSCAD CSG Format (*.csg *.CSG)", "importCSG")

param = FreeCAD.ParamGet(\
        "User parameter:BaseApp/Preferences/Mod/OpenSCAD")
openscadfilename = param.GetString('openscadexecutable')
openscadbin = openscadfilename and os.path.isfile(openscadfilename)

if openscadbin:
    FreeCAD.addImportType("OpenSCAD Format (*.scad *.SCAD)", "importCSG")
    FreeCAD.__unit_test__ += ["TestOpenSCADApp"]

FreeCAD.addExportType("OpenSCAD CSG Format (*.csg)", "exportCSG")
FreeCAD.addExportType("OpenSCAD Format (*.scad)", "exportCSG")



#! python
# (c) 2011 Werner Mayer LGPL
#

from __future__ import division  # allows floating point division from integers
import FreeCAD, Part, math
from FreeCAD import Base


class Epitrochoid:
    def __init__(self, obj):
        """Add the properties: Radius1, Radius2, Distance, Segments"""
        obj.addProperty(
            "App::PropertyLength", "Radius1", "Epitrochoid", "Interior radius", locked=True
        ).Radius1 = 60.0
        obj.addProperty(
            "App::PropertyLength", "Radius2", "Epitrochoid", "Exterior radius", locked=True
        ).Radius2 = 20.0
        obj.addProperty(
            "App::PropertyLength", "Distance", "Epitrochoid", "Distance", locked=True
        ).Distance = 10.0
        obj.addProperty(
            "App::PropertyInteger",
            "Segments",
            "Epitrochoid",
            "Number of the line segments",
            locked=True,
        ).Segments = 72
        obj.Proxy = self

    def onChanged(self, fp, prop):
        if (
            prop == "Radius1" or prop == "Radius2" or prop == "Distance" or prop == "Segments"
        ):  # if one of these is changed
            self.execute(fp)

    def execute(self, fp):  # main part of script
        steps = fp.Segments  # get value from property
        dang = math.radians(360 / steps)
        r2 = fp.Radius2
        r1 = fp.Radius1
        f1 = r1 + r2
        f2 = f1 / r2
        d = fp.Distance
        ang = 0
        z = 0

        if r2 == 0:
            raise ValueError("Exterior radius must not be zero")

        for i in range(steps):
            if i == 0:
                x1 = f1 * math.cos(ang) - d * math.cos(f2 * ang)  # coords for line startpoint
                y1 = f1 * math.sin(ang) - d * math.sin(f2 * ang)
                ang = dang
                x2 = f1 * math.cos(ang) - d * math.cos(f2 * ang)  # coords for line endpoint
                y2 = f1 * math.sin(ang) - d * math.sin(f2 * ang)
                seg = Part.makeLine((x1, y1, z), (x2, y2, z))
                wire = Part.Wire([seg])
                x1 = x2
                y1 = y2
            else:
                x2 = f1 * math.cos(ang) - d * math.cos(f2 * ang)
                y2 = f1 * math.sin(ang) - d * math.sin(f2 * ang)
                seg = Part.makeLine((x1, y1, z), (x2, y2, z))
                wire = Part.Wire([wire, seg])
                x1 = x2
                y1 = y2
            ang = ang + dang  # increment angle
        fp.Shape = wire  # result shape


def makeEpitrochoid():
    doc = FreeCAD.activeDocument()
    if doc is None:
        doc = FreeCAD.newDocument()
    epitrochoid = doc.addObject("Part::FeaturePython", "Epitrochoid")  # add object to document
    epitrochoid.Label = "Epitrochoid"
    Epitrochoid(epitrochoid)
    epitrochoid.ViewObject.Proxy = 0
    doc.recompute()


if __name__ == "__main__":  # feature will be generated after macro execution
    makeEpitrochoid()



# ***************************************************************************
# *   Copyright (c) 2019 Johannes Hartung <j.hartung@gmx.net>               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD YAML and JSON mesh reader and writer"
__author__ = "Johannes Hartung"
__url__ = "https://www.freecad.org"

## @package importYAMLJSONMesh
#  \ingroup FEM
#  \brief FreeCAD YAML and JSON Mesh reader and writer for FEM workbench

import json
import os

import FreeCAD
from FreeCAD import Console

from . import importToolsFem
from builtins import open as pyopen

has_yaml = True
try:
    import yaml
except ImportError:
    Console.PrintMessage("No YAML available (import yaml failure), yaml import/export won't work\n")
    has_yaml = False


# ****************************************************************************
# ********* generic FreeCAD import and export methods ************************
# names are fix given from FreeCAD, these methods are called from FreeCAD
# they are set in FEM modules Init.py


def open(filename):
    """called when freecad opens a file
    a FEM mesh object is created in a new document"""

    docname = os.path.splitext(os.path.basename(filename))[0]
    return insert(filename, docname)


def insert(filename, docname):
    """called when freecad wants to import a file"
    a FEM mesh object is created in a existing document"""

    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc

    import_yaml_json_mesh(filename)
    return doc


def export(objectslist, fileString):
    "called when freecad exports a file"
    if len(objectslist) != 1:
        Console.PrintError("This exporter can only export one object.\n")
        return
    obj = objectslist[0]
    if not obj.isDerivedFrom("Fem::FemMeshObject"):
        Console.PrintError("No FEM mesh object selected.\n")
        return

    write(fileString, obj.FemMesh)


# ****************************************************************************
# ********* module specific methods ******************************************
# reader:
# - a method uses a FemMesh instance, creates the FEM mesh document object and
#     returns this object
# - a method read the data from file creates FemMesh instance out of the
#     FEM mesh dictionary. This instance is returned
# - a converts the raw read data into the FEM mesh dictionary which
#     can be used to create a FemMesh instance
#
#
# writer:
# - a method directly writes a FemMesh to the mesh file


# ********* reader ***********************************************************
def import_yaml_json_mesh(fileString):
    """
    read a FemMesh from a yaml/json mesh file
    insert a FreeCAD FEM Mesh object in the ActiveDocument
    return the FEM mesh document object
    """

    mesh_name = os.path.basename(os.path.splitext(fileString)[0])

    femmesh = read(fileString)
    if femmesh:
        mesh_object = FreeCAD.ActiveDocument.addObject("Fem::FemMeshObject", mesh_name)
        mesh_object.FemMesh = femmesh

    return mesh_object


def read(fileString):
    """read a FemMesh from a yaml/json mesh file and return the FemMesh"""
    # no document object is created, just the FemMesh is returned

    fileExtension = os.path.basename(os.path.splitext(fileString)[1])

    raw_mesh_data = {}
    if fileExtension.lower() == ".meshjson" or fileExtension.lower() == ".json":
        fp = pyopen(fileString, "rt")
        raw_mesh_data = json.load(fp)
        fp.close()
    elif (
        fileExtension.lower() == ".meshyaml"
        or fileExtension.lower() == ".meshyml"
        or fileExtension.lower() == ".yaml"
        or fileExtension.lower() == ".yml"
    ) and has_yaml:
        fp = pyopen(fileString, "rt")
        raw_mesh_data = yaml.load(fp, Loader=yaml.SafeLoader)
        fp.close()
    else:
        Console.PrintError("Unknown extension, please select other importer.\n")

    Console.PrintMessage("Converting indices to integer numbers ...")
    mesh_data = convert_raw_data_to_mesh_data(raw_mesh_data)
    Console.PrintMessage("OK\n")

    return importToolsFem.make_femmesh(mesh_data)


def convert_raw_data_to_mesh_data(raw_mesh_data):
    """
    Converts raw dictionary data from JSON or YAML file to proper dict
    for importToolsFem.make_femmesh(mesh_data). This is necessary since
    JSON and YAML save dict keys as strings while make_femmesh expects
    integers.
    """

    mesh_data = {}
    for type_key, type_dict in raw_mesh_data.items():
        if type_key.lower() != "groups":
            mesh_data[type_key] = {int(k): v for (k, v) in type_dict.items()}
    return mesh_data


# ********* writer ***********************************************************
def write(fileString, fem_mesh):
    """directly write a FemMesh to a yaml/json mesh file
    fem_mesh: a FemMesh"""

    mesh_data = importToolsFem.make_dict_from_femmesh(fem_mesh)

    if fileString != "":
        fileName, fileExtension = os.path.splitext(fileString)
        if fileExtension.lower() == ".json" or fileExtension.lower() == ".meshjson":
            fp = pyopen(fileString, "wt")
            json.dump(mesh_data, fp, indent=4)
            fp.close()
        elif (
            fileExtension.lower() == ".meshyaml"
            or fileExtension.lower() == ".meshyml"
            or fileExtension.lower() == ".yaml"
            or fileExtension.lower() == ".yml"
        ) and has_yaml:
            fp = pyopen(fileString, "wt")
            yaml.safe_dump(mesh_data, fp)
            fp.close()



# ***************************************************************************
# *   Copyright (c) 2024 Werner Mayer <wmayer[at]users.sourceforge.net>     *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import unittest

import FreeCAD

""" Test transaction interaction """


class TestSketch(unittest.TestCase):
    def setUp(self):
        self.doc = FreeCAD.newDocument("PartDesignTestSketch")
        self.doc.UndoMode = True

    def testIssue17553(self):
        self.doc.openTransaction("Create box")
        box = self.doc.addObject("Part::Box", "Box")
        self.doc.commitTransaction()
        self.doc.recompute()

        self.doc.openTransaction("Create sketch")
        body = self.doc.addObject("PartDesign::Body", "Body")
        plane = self.doc.getObject("XY_Plane")
        self.doc.commitTransaction()

        self.doc.openTransaction("Rename object")
        box.Label = "Object"
        self.doc.commitTransaction()

        sketch = body.newObject("Sketcher::SketchObject", "Sketch")
        sketch.AttachmentSupport = (plane, [""])
        sketch.MapMode = "FlatFace"
        self.doc.recompute()

        self.assertEqual(sketch.InList, [body])
        self.assertEqual(sketch.OutList, [plane])
        sketch.AttachmentSupport == [(plane, ("",))]

        self.doc.undo()  # undo renaming
        self.doc.undo()  # undo body creation
        self.doc.undo()  # undo box creation

        self.doc.openTransaction("Remove sketch")
        self.doc.removeObject(sketch.Name)
        self.doc.commitTransaction()

        self.doc.undo()  # undo removal

        self.assertEqual(sketch.InList, [])
        self.assertEqual(sketch.OutList, [])
        self.assertEqual(sketch.AttachmentSupport, [])

    def testDependency(self):
        self.doc.openTransaction("Create box")
        box = self.doc.addObject("Part::Box", "Box")
        self.doc.commitTransaction()
        self.doc.recompute()

        self.doc.openTransaction("Create sketch")
        body = self.doc.addObject("PartDesign::Body", "Body")
        plane = self.doc.getObject("XY_Plane")
        self.doc.commitTransaction()

        self.doc.openTransaction("Rename object")
        box.Label = "Object"
        self.doc.commitTransaction()

        sketch = body.newObject("Sketcher::SketchObject", "Sketch")
        sketch.AttachmentSupport = (plane, [""])
        sketch.MapMode = "FlatFace"
        self.doc.recompute()

        sketch.OutList
        sketch.AttachmentSupport

        self.doc.undo()  # undo renaming
        self.doc.undo()  # undo body creation
        self.doc.undo()  # undo box creation

        self.doc.DependencyGraph

    def tearDown(self):
        FreeCAD.closeDocument("PartDesignTestSketch")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Arch Truss"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchTruss
#  \ingroup ARCH
#  \brief The Truss object and tools
#
#  This module provides tools to build Truss objects.

import math

import FreeCAD
import ArchComponent

if FreeCAD.GuiUp:
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond

rodmodes = (
    "/|/|/|",
    "/\\/\\/\\",
    "/|\\|/|\\",
)


class Truss(ArchComponent.Component):
    "The truss object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "Truss"
        self.setProperties(obj)
        obj.IfcType = "Beam"

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "TrussAngle" in pl:
            obj.addProperty(
                "App::PropertyAngle",
                "TrussAngle",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "The angle of the truss"),
                locked=True,
            )
            obj.setEditorMode("TrussAngle", 1)
        if not "SlantType" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "SlantType",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "The slant type of this truss"),
                locked=True,
            )
            obj.SlantType = ["Simple", "Double"]
        if not "Normal" in pl:
            obj.addProperty(
                "App::PropertyVector",
                "Normal",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "The normal direction of this truss"),
                locked=True,
            )
            obj.Normal = FreeCAD.Vector(0, 0, 1)
        if not "HeightStart" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "HeightStart",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "The height of the truss at the start position"),
                locked=True,
            )
            obj.HeightStart = 100
        if not "HeightEnd" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "HeightEnd",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "The height of the truss at the end position"),
                locked=True,
            )
            obj.HeightEnd = 200
        if not "StrutStartOffset" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "StrutStartOffset",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "An optional start offset for the top strut"),
                locked=True,
            )
        if not "StrutEndOffset" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "StrutEndOffset",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "An optional end offset for the top strut"),
                locked=True,
            )
        if not "StrutHeight" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "StrutHeight",
                "Truss",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The height of the main top and bottom elements of the truss"
                ),
                locked=True,
            )
            obj.StrutHeight = 10
        if not "StrutWidth" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "StrutWidth",
                "Truss",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The width of the main top and bottom elements of the truss"
                ),
                locked=True,
            )
            obj.StrutWidth = 5
        if not "RodType" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "RodType",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "The type of the middle element of the truss"),
                locked=True,
            )
            obj.RodType = ["Round", "Square"]
        if not "RodDirection" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "RodDirection",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "The direction of the rods"),
                locked=True,
            )
            obj.RodDirection = ["Forward", "Backward"]
        if not "RodSize" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "RodSize",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "The diameter or side of the rods"),
                locked=True,
            )
            obj.RodSize = 2
        if not "RodSections" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "RodSections",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "The number of rod sections"),
                locked=True,
            )
            obj.RodSections = 3
        if not "RodEnd" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "RodEnd",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "If the truss has a rod at its endpoint or not"),
                locked=True,
            )
        if not "RodMode" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "RodMode",
                "Truss",
                QT_TRANSLATE_NOOP("App::Property", "How to draw the rods"),
                locked=True,
            )
            obj.RodMode = rodmodes

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

    def loads(self, state):

        self.Type = "Truss"

    def onChanged(self, obj, prop):

        ArchComponent.Component.onChanged(self, obj, prop)

    def execute(self, obj):

        if self.clone(obj):
            return
        if not self.ensureBase(obj):
            return

        import Part

        pl = obj.Placement

        # test properties
        if not obj.StrutWidth.Value or not obj.StrutHeight.Value:
            FreeCAD.Console.PrintLog(obj.Label + ": Invalid strut size\n")
            return
        if not obj.HeightStart.Value or not obj.HeightEnd.Value:
            FreeCAD.Console.PrintLog(obj.Label + ": Invalid truss heights\n")
            return
        if not obj.RodSections or not obj.RodSize.Value:
            FreeCAD.Console.PrintLog(obj.Label + ": Invalid rod config\n")
            return
        if not obj.Base:
            FreeCAD.Console.PrintLog(obj.Label + ": No base\n")
            return
        if (not hasattr(obj.Base, "Shape")) or (len(obj.Base.Shape.Vertexes) > 2):
            FreeCAD.Console.PrintLog(obj.Label + ": No base edge\n")
            return

        # baseline
        v0 = obj.Base.Shape.Vertexes[0].Point
        v1 = obj.Base.Shape.Vertexes[-1].Point

        if obj.SlantType == "Simple":
            topstrut, bottomstrut, rods, angle = self.makeTruss(obj, v0, v1)
        else:
            v3 = v0.add((v1.sub(v0)).multiply(0.5))
            topstrut1, bottomstrut1, rods1, angle = self.makeTruss(obj, v0, v3)
            topstrut2, bottomstrut2, rods2, angle = self.makeTruss(obj, v1, v3)
            topstrut = topstrut1.fuse(topstrut2)
            topstrut = topstrut.removeSplitter()
            bottomstrut = bottomstrut1.fuse(bottomstrut2)
            bottomstrut = bottomstrut.removeSplitter()
            if obj.RodEnd:
                rods2 = rods2[:-1]
            rods = rods1 + rods2

        # mount shape
        shape = Part.makeCompound([topstrut, bottomstrut] + rods)
        shape = self.processSubShapes(obj, shape, pl)
        self.applyShape(obj, shape, pl)
        obj.TrussAngle = angle

    def makeTruss(self, obj, v0, v1):

        import Part

        # get normal direction
        normal = obj.Normal
        if not normal.Length:
            normal = FreeCAD.Vector(0, 0, 1)

        # create base profile
        maindir = v1.sub(v0)
        sidedir = normal.cross(maindir)
        if not sidedir.Length:
            FreeCAD.Console.PrintLog(obj.Label + ": normal and base are parallel\n")
            return
        sidedir.normalize()
        p0 = v0.add(FreeCAD.Vector(sidedir).negative().multiply(obj.StrutWidth.Value / 2))
        p1 = p0.add(FreeCAD.Vector(sidedir).multiply(obj.StrutWidth.Value / 2).multiply(2))
        p2 = p1.add(FreeCAD.Vector(normal).multiply(obj.StrutHeight))
        p3 = p0.add(FreeCAD.Vector(normal).multiply(obj.StrutHeight))
        trussprofile = Part.Face(Part.makePolygon([p0, p1, p2, p3, p0]))

        # create bottom strut
        bottomstrut = trussprofile.extrude(maindir)

        # create top strut
        v2 = v0.add(FreeCAD.Vector(normal).multiply(obj.HeightStart.Value))
        v3 = v1.add(FreeCAD.Vector(normal).multiply(obj.HeightEnd.Value))
        topdir = v3.sub(v2)
        if obj.StrutStartOffset.Value:
            v2f = v2.add((v2.sub(v3)).normalize().multiply(obj.StrutStartOffset.Value))
        else:
            v2f = v2
        if obj.StrutEndOffset.Value:
            v3f = v3.add((v3.sub(v2)).normalize().multiply(obj.StrutEndOffset.Value))
        else:
            v3f = v3
        offtopdir = v3f.sub(v2f)
        topstrut = trussprofile.extrude(offtopdir)
        topstrut.translate(v2f.sub(v0))
        topstrut.translate(FreeCAD.Vector(normal).multiply(-obj.StrutHeight.Value))
        angle = math.degrees(topdir.getAngle(maindir))

        # create rod profile on the XY plane
        if obj.RodType == "Round":
            rodprofile = Part.Face(Part.Wire([Part.makeCircle(obj.RodSize / 2)]))
        else:
            rodprofile = Part.Face(
                Part.makePlane(
                    obj.RodSize, obj.RodSize, FreeCAD.Vector(-obj.RodSize / 2, -obj.RodSize / 2, 0)
                )
            )

        # create rods
        rods = []
        bottomrodstart = v0.add(FreeCAD.Vector(normal).multiply(obj.StrutHeight.Value / 2))
        toprodstart = v2.add(FreeCAD.Vector(normal).multiply(obj.StrutHeight.Value / 2).negative())
        bottomrodvec = FreeCAD.Vector(maindir).multiply(1 / obj.RodSections)
        toprodvec = FreeCAD.Vector(topdir).multiply(1 / obj.RodSections)
        bottomrodpos = [bottomrodstart]
        toprodpos = [toprodstart]

        if obj.RodMode == rodmodes[0]:
            # /|/|/|
            for i in range(1, obj.RodSections + 1):
                if (i > 1) or (obj.StrutStartOffset.Value >= 0):
                    # do not add first vert rod if negative offset
                    rods.append(self.makeRod(rodprofile, bottomrodpos[-1], toprodpos[-1]))
                bottomrodpos.append(bottomrodpos[-1].add(bottomrodvec))
                toprodpos.append(toprodpos[-1].add(toprodvec))
                if obj.RodDirection == "Forward":
                    rods.append(self.makeRod(rodprofile, bottomrodpos[-2], toprodpos[-1]))
                else:
                    rods.append(self.makeRod(rodprofile, bottomrodpos[-1], toprodpos[-2]))

        elif obj.RodMode == rodmodes[1]:
            # /\/\/\
            fw = True
            for i in range(1, obj.RodSections + 1):
                bottomrodpos.append(bottomrodpos[-1].add(bottomrodvec))
                toprodpos.append(toprodpos[-1].add(toprodvec))
                if obj.RodDirection == "Forward":
                    if fw:
                        rods.append(self.makeRod(rodprofile, bottomrodpos[-2], toprodpos[-1]))
                    else:
                        rods.append(self.makeRod(rodprofile, bottomrodpos[-1], toprodpos[-2]))
                else:
                    if fw:
                        rods.append(self.makeRod(rodprofile, bottomrodpos[-1], toprodpos[-2]))
                    else:
                        rods.append(self.makeRod(rodprofile, bottomrodpos[-2], toprodpos[-1]))
                fw = not fw

        elif obj.RodMode == rodmodes[2]:
            # /|\|/|\
            fw = True
            for i in range(1, obj.RodSections + 1):
                if (i > 1) or (obj.StrutStartOffset.Value >= 0):
                    # do not add first vert rod if negative offset
                    rods.append(self.makeRod(rodprofile, bottomrodpos[-1], toprodpos[-1]))
                bottomrodpos.append(bottomrodpos[-1].add(bottomrodvec))
                toprodpos.append(toprodpos[-1].add(toprodvec))
                if obj.RodDirection == "Forward":
                    if fw:
                        rods.append(self.makeRod(rodprofile, bottomrodpos[-2], toprodpos[-1]))
                    else:
                        rods.append(self.makeRod(rodprofile, bottomrodpos[-1], toprodpos[-2]))
                else:
                    if fw:
                        rods.append(self.makeRod(rodprofile, bottomrodpos[-1], toprodpos[-2]))
                    else:
                        rods.append(self.makeRod(rodprofile, bottomrodpos[-2], toprodpos[-1]))
                fw = not fw

        # add end rod
        if obj.RodEnd:
            rods.append(self.makeRod(rodprofile, bottomrodpos[-1], toprodpos[-1]))

        # trim rods
        rods = [rod.cut(topstrut).cut(bottomstrut) for rod in rods]

        return topstrut, bottomstrut, rods, angle

    def makeRod(self, profile, p1, p2):
        """makes a rod by extruding profile between p1 and p2"""

        rot = FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), p2.sub(p1))
        rod = profile.copy().translate(p1)
        rod = rod.rotate(p1, rot.Axis, math.degrees(rot.Angle))
        rod = rod.extrude(p2.sub(p1))
        return rod


class ViewProviderTruss(ArchComponent.ViewProviderComponent):
    "A View Provider for the Truss object"

    def __init__(self, vobj):

        ArchComponent.ViewProviderComponent.__init__(self, vobj)

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_Truss_Tree.svg"



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM StaticCurrent Elmer writer"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from FreeCAD import Units
from .. import sifio


class SCwriter:

    def __init__(self, writer, solver):
        self.write = writer
        self.solver = solver

    def getStaticCurrentSolver(self, equation):
        # output the equation parameters
        s = self.write.createNonlinearSolver(equation)
        s["Equation"] = "Stat Current Solver"
        s["Procedure"] = sifio.FileAttr("StatCurrentSolve/StatCurrentSolver")
        s["Variable"] = self.write.getUniqueVarName("Potential")
        s["Variable DOFs"] = 1
        s["Calculate Volume Current"] = equation.CalculateVolumeCurrent
        s["Calculate Joule Heating"] = equation.CalculateJouleHeating
        s["Constant Weights"] = equation.ConstantWeights
        s["Calculate Nodal Heating"] = equation.CalculateNodalHeating
        if equation.PowerControl:
            s["Power Control"] = equation.Power.getValueAs("W").Value
        if equation.CurrentControl:
            s["Current Control"] = equation.Current.getValueAs("A").Value
        s["Exec Solver"] = "Always"
        s["Optimize Bandwidth"] = True
        s["Stabilize"] = equation.Stabilize

        return s

    def handleStaticCurrentConstants(self):
        pass

    def handleStaticCurrentMaterial(self, bodies):
        for obj in self.write.getMember("App::MaterialObject"):
            m = obj.Material
            refs = obj.References[0][1] if obj.References else self.write.getAllBodies()
            for name in (n for n in refs if n in bodies):
                self.write.material(name, "Name", m["Name"])
                if "ElectricalConductivity" in m:
                    self.write.material(
                        name,
                        "Electric Conductivity",
                        Units.Quantity(m["ElectricalConductivity"]).getValueAs("S/m").Value,
                    )

    def handleStaticCurrentBndConditions(self):
        for obj in self.write.getMember("Fem::ConstraintElectrostaticPotential"):
            if obj.References:
                for name in obj.References[0][1]:
                    # output the FreeCAD label as comment
                    if obj.Label:
                        self.write.boundary(name, "! FreeCAD Name", obj.Label)
                    if obj.BoundaryCondition == "Dirichlet":
                        if obj.PotentialEnabled:
                            self.write.boundary(name, "Current Density BC", False)
                            self.write.boundary(
                                name, "Potential", obj.Potential.getValueAs("V").Value
                            )
                self.write.handled(obj)

        for obj in self.write.getMember("Fem::ConstraintCurrentDensity"):
            if obj.References:
                for name in obj.References[0][1]:
                    # output the FreeCAD label as comment
                    if obj.Label:
                        self.write.boundary(name, "! FreeCAD Name", obj.Label)
                    if obj.Mode == "Normal":
                        self.write.boundary(name, "Current Density BC", True)
                        self.write.boundary(
                            name,
                            "Current Density",
                            obj.NormalCurrentDensity_re.getValueAs("A/m^2").Value,
                        )

                self.write.handled(obj)

    def handleStaticCurrentBodyForces(self, bodies, equation):
        for name in bodies:
            self.write.bodyForce(name, "Joule Heat", equation.HeatSource)


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# /***************************************************************************
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# **************************************************************************

__title__ = "FreeCAD FEM calculix constraint body heat source"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

import FreeCAD
import itertools


def get_analysis_types():
    return ["thermomech"]


def get_sets_name():
    return "constraints_bodyheatsource_element_sets"


def get_constraint_title():
    return "Body Heat Source Constraints"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_meshdata_constraint(f, femobj, bodyheatsource_obj, ccxwriter):

    ref_iter = itertools.chain(
        *[itertools.product([i[0]], i[1]) for i in bodyheatsource_obj.References]
    )
    ref_iter = tuple(ref_iter)
    for refs, surf, is_sub_el in femobj["BodyHeatSourceElements"]:
        feat, (sub,) = refs
        index = ref_iter.index((feat, sub))
        f.write(f"** {bodyheatsource_obj.Name}.{feat.Name}.{sub}\n")
        f.write(f"*ELSET,ELSET={bodyheatsource_obj.Name}_{index}\n")
        if not is_sub_el:
            for elem in surf:
                f.write(f"{elem},\n")
        f.write("\n")


def write_constraint(f, femobj, bodyheatsource_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module
    ref_iter = itertools.chain(
        *[itertools.product([i[0]], i[1]) for i in bodyheatsource_obj.References]
    )
    for index, (ref_feat, ref_sub_obj) in enumerate(ref_iter):

        # search referenced material
        mat = _search_member_property(
            ccxwriter.member.mats_linear, (ref_feat, ref_sub_obj), "Material"
        )
        density = FreeCAD.Units.Quantity(mat["Density"])

        # get data from the bodyheatsource_obj (DissipationRate is in power per unit mass)
        if bodyheatsource_obj.Mode == "Dissipation Rate":
            heat = bodyheatsource_obj.DissipationRate * density
        elif bodyheatsource_obj.Mode == "Total Power":
            sh = ref_feat.getSubObject(ref_sub_obj)
            volume = sh.Mass
            # if shape is face, get thickness
            if sh.ShapeType == "Face":
                thick = _search_member_property(
                    ccxwriter.member.geos_shellthickness, (ref_feat, ref_sub_obj), "Thickness"
                )
                volume *= thick.getValueAs("mm")
            heat = bodyheatsource_obj.TotalPower / FreeCAD.Units.Quantity(volume, "mm^3")

        # write to file
        if bodyheatsource_obj.EnableAmplitude:
            f.write(f"*DFLUX, AMPLITUDE={bodyheatsource_obj.Name}\n")
        else:
            f.write("*DFLUX\n")
        elset_name = f"{bodyheatsource_obj.Name}_{index}"
        f.write("{},BF,{:.13G}\n".format(elset_name, heat.getValueAs("t/(mm*s^3)").Value))
        f.write("\n")


def _search_member_property(it, ref, prop_name):
    value = None
    for member in it:
        member_ref = [
            *itertools.chain(
                *[itertools.product([i[0]], i[1]) for i in member["Object"].References]
            )
        ]
        if ref in member_ref:
            value = member["Object"].getPropertyByName(prop_name)
            break

    if not value:
        # search membererial without references
        for member in it:
            if not member["Object"].References:
                value = member["Object"].getPropertyByName(prop_name)
                break

    return value
